import datetime
import math
import os
import shutil
import warnings
from enum import Enum
from pathlib import Path
from typing import Optional, Union
from urllib.request import urlretrieve

import cht_observations.observation_stations as obs
import geopandas as gpd
import numpy as np
import pandas as pd
import rioxarray as rxr
import tomli
import tomli_w
import xarray as xr
from hydromt_fiat import FiatModel as HydromtFiatModel
from hydromt_fiat.data_apis.open_street_maps import get_buildings_from_osm
from hydromt_sfincs import SfincsModel as HydromtSfincsModel
from pydantic import BaseModel, Field
from shapely import MultiLineString, MultiPolygon, Polygon

from flood_adapt.adapter.fiat_adapter import _FIAT_COLUMNS
from flood_adapt.config.fiat import (
    AggregationModel,
    BenefitsModel,
    BFEModel,
    EquityModel,
    FiatConfigModel,
    FiatModel,
    RiskModel,
    SVIModel,
)
from flood_adapt.config.gui import (
    AggregationDmgLayer,
    BenefitsLayer,
    FloodMapLayer,
    FootprintsDmgLayer,
    GuiModel,
    GuiUnitModel,
    OutputLayers,
    PlottingModel,
    SyntheticTideModel,
    VisualizationLayers,
)
from flood_adapt.config.sfincs import (
    Cstype,
    CycloneTrackDatabaseModel,
    DatumModel,
    DemModel,
    FloodmapType,
    FloodModel,
    ObsPointModel,
    RiverModel,
    SCSModel,
    SfincsConfigModel,
    SfincsModel,
    SlrScenariosModel,
    WaterlevelReferenceModel,
)
from flood_adapt.config.site import (
    Site,
    StandardObjectModel,
)
from flood_adapt.dbs_classes.database import Database
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.utils import modified_environ
from flood_adapt.objects.events.event_set import EventSet
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.tide_gauge import (
    TideGauge,
    TideGaugeSource,
)
from flood_adapt.objects.projections.projections import (
    PhysicalProjection,
    Projection,
    SocioEconomicChange,
)
from flood_adapt.objects.strategies.strategies import Strategy


def path_check(str_path: str, config_path: Optional[Path] = None) -> str:
    """
    Check if the given path is absolute and return the absolute path.

    Args:
        path (str): The path to be checked.

    Returns
    -------
        str: The absolute path.

    Raises
    ------
        ValueError: If the path is not absolute and no config_path is provided.
    """
    path = Path(str_path)
    if not path.is_absolute():
        if config_path is not None:
            path = Path(config_path).parent.joinpath(path).resolve()
        else:
            raise ValueError(f"Value '{path}' should be an absolute path.")
    return path.as_posix()


class SpatialJoinModel(BaseModel):
    """
    Represents a spatial join model.

    Attributes
    ----------
    name (Optional[str]): The name of the model (optional).
    file (str): The file associated with the model.
    field_name (str): The field name used for the spatial join.
    """

    name: Optional[str] = None
    file: str
    field_name: str


class UnitSystems(str, Enum):
    """The `UnitSystems` class is an enumeration that represents the accepted values for the `metric_system` field.

    It provides two options: `imperial` and `metric`.

    Attributes
    ----------
        imperial (str): Represents the imperial unit system.
        metric (str): Represents the metric unit system.
    """

    imperial = "imperial"
    metric = "metric"


class FootprintsOptions(str, Enum):
    OSM = "OSM"


class Basins(str, Enum):
    """
    Enumeration class representing different basins.

    Each basin is represented by a string value.

    Attributes
    ----------
        NA (str): North Atlantic
        SA (str): South Atlantic
        EP (str): Eastern North Pacific (which includes the Central Pacific region)
        WP (str): Western North Pacific
        SP (str): South Pacific
        SI (str): South Indian
        NI (str): North Indian
    """

    NA = "NA"
    SA = "SA"
    EP = "EP"
    WP = "WP"
    SP = "SP"
    SI = "SI"
    NI = "NI"


class GuiConfigModel(BaseModel):
    """
    Represents a GUI model for for FloodAdapt.

    Attributes
    ----------
        max_flood_depth (float): The last visualization bin will be ">value".
        max_aggr_dmg (float): The last visualization bin will be ">value".
        max_footprint_dmg (float): The last visualization bin will be ">value".
        max_benefits (float): The last visualization bin will be ">value".
    """

    max_flood_depth: float
    max_aggr_dmg: float
    max_footprint_dmg: float
    max_benefits: float


class SviConfigModel(SpatialJoinModel):
    """
    Represents a model for the Social Vulnerability Index (SVI).

    Attributes
    ----------
        threshold (float): The threshold value for the SVI model to specify vulnerability.
    """

    threshold: float


class Point(BaseModel):
    lat: float
    lon: float


class TideGaugeConfigModel(BaseModel):
    """
    Represents a tide gauge model.

    Attributes
    ----------
        source (str): The source of the tide gauge data.
        file (Optional[str]): The file associated with the tide gauge data (default: None).
        max_distance (Optional[float]): The maximum distance (default: None).
        ref (str): The reference name. Should be defined in the water level references.
    """

    source: TideGaugeSource
    description: str = ""
    ref: Optional[str] = None
    id: Optional[int] = None
    lon: Optional[float] = None
    lat: Optional[float] = None
    file: Optional[str] = None
    max_distance: Optional[us.UnitfulLength] = None


class SviModel(SpatialJoinModel):
    """
    Represents a model for the Social Vulnerability Index (SVI).

    Attributes
    ----------
        threshold (float): The threshold value for the SVI model to specify vulnerability.
    """

    threshold: float


class ConfigModel(BaseModel):
    """
    Represents the configuration model for FloodAdapt.

    Attributes
    ----------
    name : str
        The name of the site.
    description : Optional[str], default ""
        The description of the site.
    database_path : Optional[str], default None
        The path to the database where all the sites are located.
    sfincs : str
        The SFINCS model path.
    sfincs_offshore : Optional[str], default None
        The offshore SFINCS model path.
    fiat : str
        The FIAT model path.
    unit_system : UnitSystems
        The unit system.
    gui : GuiModel
        The GUI model representing scaling values for the layers.
    building_footprints : Optional[SpatialJoinModel], default None
        The building footprints model.
    slr_scenarios : Optional[SlrModelDef], default SlrModelDef()
        The sea level rise model.
    tide_gauge : Optional[TideGaugeConfigModel], default None
        The tide gauge model.
    bfe : Optional[SpatialJoinModel], default None
        The BFE model.
    svi : Optional[SviModel], default None
        The SVI model.
    road_width : Optional[float], default 2
        The road width in meters.
    cyclones : Optional[bool], default True
        Indicates if cyclones are enabled.
    cyclone_basin : Optional[Basins], default None
        The cyclone basin.
    obs_point : Optional[list[ObsPointModel]], default None
        The list of observation point models.
    probabilistic_set : Optional[str], default None
        The probabilistic set path.
    infographics : Optional[bool], default True
        Indicates if infographics are enabled.
    """

    # General
    name: str = Field(..., min_length=1, pattern='^[^<>:"/\\\\|?* ]*$')
    description: Optional[str] = None
    database_path: Optional[str] = None
    unit_system: UnitSystems
    gui: GuiConfigModel
    infographics: Optional[bool] = True

    # FIAT
    fiat: str
    aggregation_areas: Optional[list[SpatialJoinModel]] = None
    building_footprints: Optional[SpatialJoinModel | FootprintsOptions] = (
        FootprintsOptions.OSM
    )
    fiat_buildings_name: Optional[str] = "buildings"
    fiat_roads_name: Optional[str] = "roads"
    bfe: Optional[SpatialJoinModel] = None
    svi: Optional[SviConfigModel] = None
    road_width: Optional[float] = 5
    return_periods: list[int] = Field(default_factory=list)

    # SFINCS
    references: WaterlevelReferenceModel = WaterlevelReferenceModel(
        reference="MSL",
        datums=[
            DatumModel(
                name="MSL",
                height=us.UnitfulLength(value=0.0, units=us.UnitTypesLength.meters),
            ),
        ],
    )

    sfincs_overland: FloodModel
    sfincs_offshore: Optional[FloodModel] = None
    dem: Optional[DemModel] = None

    excluded_datums: list[str] = Field(default_factory=list)

    slr_scenarios: Optional[SlrScenariosModel] = None
    scs: Optional[SCSModel] = None
    tide_gauge: Optional[TideGaugeConfigModel] = None
    cyclones: Optional[bool] = True
    cyclone_basin: Optional[Basins] = None
    obs_point: Optional[list[ObsPointModel]] = None
    probabilistic_set: Optional[str] = None

    @staticmethod
    def read(toml_path: Path) -> "ConfigModel":
        """
        Read a configuration file and returns the validated attributes.

        Args:
            config (str): The path to the configuration file.

        Returns
        -------
            ConfigModel: The validated attributes from the configuration file.
        """
        with open(toml_path, mode="rb") as fp:
            toml = tomli.load(fp)
        config = ConfigModel.model_validate(toml)

        # check if database path is provided and use config_file path if not
        if config.database_path is None:
            dbs_path = Path(toml_path).parent / "Database"
            if not dbs_path.exists():
                dbs_path.mkdir(parents=True)
            config.database_path = dbs_path.as_posix()
        # check if paths are relative to the config file and make them absolute
        config.database_path = path_check(config.database_path, toml_path)
        config.fiat = path_check(config.fiat, toml_path)
        config.sfincs_overland.name = path_check(config.sfincs_overland.name, toml_path)
        if config.sfincs_offshore:
            config.sfincs_offshore.name = path_check(
                config.sfincs_offshore.name, toml_path
            )
        if isinstance(config.building_footprints, SpatialJoinModel):
            config.building_footprints.file = path_check(
                config.building_footprints.file, toml_path
            )
        if config.tide_gauge and config.tide_gauge.file:
            config.tide_gauge.file = path_check(config.tide_gauge.file, toml_path)
        if config.svi:
            config.svi.file = path_check(config.svi.file, toml_path)
        if config.bfe:
            config.bfe.file = path_check(config.bfe.file, toml_path)
        if config.slr_scenarios:
            config.slr_scenarios.file = path_check(config.slr_scenarios.file, toml_path)
        if config.probabilistic_set:
            config.probabilistic_set = path_check(config.probabilistic_set, toml_path)
        if config.aggregation_areas:
            for aggr in config.aggregation_areas:
                aggr.file = path_check(aggr.file, toml_path)

        return config


class DatabaseBuilder:
    logger = FloodAdaptLogging.getLogger("DatabaseBuilder")

    _has_roads: bool = False
    _aggregation_areas: Optional[list] = None
    _probabilistic_set_name: Optional[str] = None

    def __init__(self, config: ConfigModel, overwrite: bool = True):
        self.config = config

        # Set database root
        if config.database_path:
            self.root = Path(config.database_path).joinpath(self.config.name)
        else:
            raise ValueError(
                "Database path is not provided. Please provide a path using the 'database_path' attribute."
            )

        # Read info that needs to be used to create other models
        self.unit_system = self.create_default_units()

        # Read info that needs to be updated with other model info
        self.water_level_references = self.config.references

    @property
    def static_path(self) -> Path:
        return self.root / "static"

    def build(self, overwrite: bool = False) -> None:
        # Check if database already exists
        if self.root.exists() and not overwrite:
            raise ValueError(
                f"There is already a Database folder in '{self.root.as_posix()}'."
            )
        if self.root.exists() and overwrite:
            shutil.rmtree(self.root)
            warnings.warn(
                f"There is already a Database folder in '{self.root.as_posix()}, which will be overwritten'."
            )
        # Create database folder
        self.root.mkdir(parents=True)

        with FloodAdaptLogging.to_file(
            file_path=self.root.joinpath("database_builder.log")
        ):
            self.logger.info(
                f"Creating a FloodAdapt database in '{self.root.as_posix()}'"
            )

            # Make folder structure and read models
            self.setup()

            # Prepare site configuration
            site = self.create_site_config()
            site.save(self.static_path / "config" / "site.toml")

            # Add infometric and infographic configurations
            self.create_infometrics()

            # Save standard objects
            self.create_standard_objects()

            # Save log file
            self.logger.info("FloodAdapt database creation finished!")

    def setup(self) -> None:
        # Create the models
        self.make_folder_structure()

        # Read user models and copy to templates
        self.read_template_fiat_model()
        self.read_template_sfincs_overland_model()
        self.read_template_sfincs_offshore_model()

    def set_standard_objects(self):
        # Define name and create object
        self._no_measures_strategy_name = "no_measures"
        self._current_projection_name = "current"
        if self._probabilistic_set_name is not None:
            event_list = [self._probabilistic_set_name]
        else:
            event_list = []
        std_obj = StandardObjectModel(
            events=event_list,
            projections=[self._current_projection_name],
            strategies=[self._no_measures_strategy_name],
        )
        return std_obj

    def create_standard_objects(self):
        with modified_environ(
            DATABASE_ROOT=str(self.root.parent),
            DATABASE_NAME=self.root.name,
        ):
            self.logger.info(
                "Creating `no measures` strategy and `current` projection."
            )
            # Create database instance
            db = Database(self.root.parent, self.config.name)
            # Create no measures strategy
            strategy = Strategy(
                name=self._no_measures_strategy_name,
                measures=[],
            )
            db.strategies.save(strategy)
            # Create current projection
            projection = Projection(
                name=self._current_projection_name,
                physical_projection=PhysicalProjection(),
                socio_economic_change=SocioEconomicChange(),
            )
            db.projections.save(projection)
            # Check prob set
            if self._probabilistic_set_name is not None:
                path_toml = (
                    db.input_path
                    / "events"
                    / self._probabilistic_set_name
                    / f"{self._probabilistic_set_name}.toml"
                )
                try:
                    EventSet.load_file(path_toml)
                except Exception as e:
                    raise ValueError(
                        f"Provided probabilistic event set '{self._probabilistic_set_name}' is not valid. Error: {e}"
                    )

    ### TEMPLATE READERS ###
    def read_template_fiat_model(self):
        user_provided = self._check_exists_and_absolute(self.config.fiat)

        # Read config model
        HydromtFiatModel(root=str(user_provided), mode="r+").read()

        # Success, so copy to db and read again
        location_in_db = self.static_path / "templates" / "fiat"
        if location_in_db.exists():
            shutil.rmtree(location_in_db)
        shutil.copytree(user_provided, location_in_db)
        in_db = HydromtFiatModel(root=str(location_in_db), mode="r+")
        in_db.read()
        # Add check to make sure the geoms are correct
        # TODO this should be handled in hydromt-FIAT
        no_geoms = len(
            [name for name in in_db.config["exposure"]["geom"].keys() if "file" in name]
        )
        in_db.exposure.exposure_geoms = in_db.exposure.exposure_geoms[:no_geoms]
        in_db.exposure._geom_names = in_db.exposure._geom_names[:no_geoms]

        # Make sure that a region polygon is included
        if "region" not in in_db.geoms:
            gdf = in_db.exposure.get_full_gdf(in_db.exposure.exposure_db)
            # Combine all geometries into a single geometry
            merged_geometry = gdf.unary_union

            # If the result is not a polygon, you can create a convex hull
            if not isinstance(merged_geometry, Polygon):
                merged_geometry = merged_geometry.convex_hull
            # Create a new GeoDataFrame with the resulting polygon
            in_db.geoms["region"] = gpd.GeoDataFrame(
                geometry=[merged_geometry], crs=gdf.crs
            )

        self.fiat_model = in_db

    def read_template_sfincs_overland_model(self):
        user_provided = self._check_exists_and_absolute(
            self.config.sfincs_overland.name
        )
        user_model = HydromtSfincsModel(root=str(user_provided), mode="r")
        user_model.read()
        if user_model.crs is None:
            raise ValueError("CRS is not defined in the SFINCS model.")

        location_in_db = self.static_path / "templates" / "overland"
        if location_in_db.exists():
            shutil.rmtree(location_in_db)
        shutil.copytree(user_provided, location_in_db)
        in_db = HydromtSfincsModel(root=str(location_in_db), mode="r+")
        in_db.read()
        self.sfincs_overland_model = in_db

    def read_template_sfincs_offshore_model(self):
        if self.config.sfincs_offshore is None:
            self.sfincs_offshore_model = None
            return
        user_provided = self._check_exists_and_absolute(
            self.config.sfincs_offshore.name
        )
        user_model = HydromtSfincsModel(root=str(user_provided), mode="r+")
        user_model.read()
        if user_model.crs is None:
            raise ValueError("CRS is not defined in the SFINCS model.")
        epsg = user_model.crs.to_epsg()

        location_in_db = self.static_path / "templates" / "offshore"
        if location_in_db.exists():
            shutil.rmtree(location_in_db)
        shutil.copytree(user_provided, location_in_db)
        in_db = HydromtSfincsModel(str(location_in_db), mode="r+")
        in_db.read(epsg=epsg)
        self.sfincs_offshore_model = in_db

    ### FIAT ###
    def create_fiat_model(self) -> FiatModel:
        fiat = FiatModel(
            config=self.create_fiat_config(),
            benefits=self.create_benefit_config(),
            risk=self.create_risk_model(),
        )
        return fiat

    def create_risk_model(self) -> Optional[RiskModel]:
        # Check if return periods are provided
        if not self.config.return_periods:
            if self._probabilistic_set_name:
                risk = RiskModel()
                self.logger.warning(
                    f"No return periods provided, but a probabilistic set is available. Using default return periods {risk.return_periods}."
                )
                return risk
            else:
                self.logger.warning(
                    "No return periods provided and no probabilistic set available. Risk calculations will not be performed."
                )
                return None
        else:
            risk = RiskModel(return_periods=self.config.return_periods)
        return risk

    def create_benefit_config(self) -> Optional[BenefitsModel]:
        if self._probabilistic_set_name is None:
            self.logger.warning(
                "No probabilistic set found in the config, benefits will not be available."
            )
            return None
        return BenefitsModel(
            current_year=datetime.datetime.now().year,
            current_projection="current",
            baseline_strategy="no_measures",
            event_set=self._probabilistic_set_name,
        )

    def create_fiat_config(self) -> FiatConfigModel:
        # Make sure only csv objects have geometries
        for i, geoms in enumerate(self.fiat_model.exposure.exposure_geoms):
            keep = geoms[_FIAT_COLUMNS.object_id].isin(
                self.fiat_model.exposure.exposure_db[_FIAT_COLUMNS.object_id]
            )
            geoms = geoms[keep].reset_index(drop=True)
            self.fiat_model.exposure.exposure_geoms[i] = geoms

        footprints = self.create_footprints()
        if footprints is not None:
            footprints = footprints.as_posix()

        # Clip hazard and reset buildings # TODO use hydromt-FIAT instead
        if not self.fiat_model.region.empty:
            self._clip_hazard_extend()

        # Store result for possible future use in create_infographics
        self._aggregation_areas = self.create_aggregation_areas()

        roads_gpkg = self.create_roads()
        non_building_names = []
        if roads_gpkg is not None:
            non_building_names.append("road")

        # Update elevations
        self.update_fiat_elevation()

        self._svi = self.create_svi()

        config = FiatConfigModel(
            exposure_crs=self.fiat_model.exposure.crs,
            floodmap_type=self.read_floodmap_type(),
            bfe=self.create_bfe(),
            non_building_names=non_building_names,
            damage_unit=self.read_damage_unit(),
            building_footprints=footprints,
            roads_file_name=roads_gpkg,
            new_development_file_name=self.create_new_developments(),
            save_simulation=False,  # TODO
            infographics=self.config.infographics,
            aggregation=self._aggregation_areas,
            svi=self._svi,
        )

        # Update output geoms names
        output_geom = {}
        counter = 0
        for key in self.fiat_model.config["exposure"]["geom"].keys():
            if "file" in key:
                counter += 1
                output_geom[f"name{counter}"] = Path(
                    self.fiat_model.config["exposure"]["geom"][key]
                ).name
        self.fiat_model.config["output"]["geom"] = output_geom
        # Update FIAT model with the new config
        self.fiat_model.write()

        return config

    def update_fiat_elevation(self):
        """
        Update the ground elevations of FIAT objects based on the SFINCS ground elevation map.

        This method reads the DEM file and the exposure CSV file, and updates the ground elevations
        of the FIAT objects (roads and buildings) based on the nearest elevation values from the DEM.
        """
        dem_file = self._dem_path
        # TODO resolve issue with double geometries in hydromt-FIAT and use update_ground_elevation method instead
        # self.fiat_model.update_ground_elevation(dem_file, grnd_elev_unit="meters")
        self.logger.info(
            "Updating FIAT objects ground elevations from SFINCS ground elevation map."
        )
        SFINCS_units = us.UnitfulLength(
            value=1.0, units=us.UnitTypesLength.meters
        )  # SFINCS is always in meters
        FIAT_units = self.unit_system.default_length_units
        conversion_factor = SFINCS_units.convert(FIAT_units)

        if not math.isclose(conversion_factor, 1):
            self.logger.info(
                f"Ground elevation for FIAT objects is in '{FIAT_units}', while SFINCS ground elevation is in 'meters'. Values in the exposure csv will be converted by a factor of {conversion_factor}"
            )

        exposure = self.fiat_model.exposure.exposure_db
        dem = rxr.open_rasterio(dem_file)
        # TODO make sure only fiat_model object changes take place!
        if self.config.fiat_roads_name in self.fiat_model.exposure.geom_names:
            roads = self.fiat_model.exposure.exposure_geoms[
                self._get_fiat_road_index()
            ].to_crs(dem.spatial_ref.crs_wkt)
            roads["centroid"] = roads.geometry.centroid  # get centroids

            x_points = xr.DataArray(roads["centroid"].x, dims="points")
            y_points = xr.DataArray(roads["centroid"].y, dims="points")
            roads["elev"] = (
                dem.sel(x=x_points, y=y_points, band=1, method="nearest").to_numpy()
                * conversion_factor
            )

            exposure.loc[
                exposure[_FIAT_COLUMNS.primary_object_type] == "road",
                _FIAT_COLUMNS.ground_floor_height,
            ] = 0
            exposure = exposure.merge(
                roads[[_FIAT_COLUMNS.object_id, "elev"]],
                on=_FIAT_COLUMNS.object_id,
                how="left",
            )
            exposure.loc[
                exposure[_FIAT_COLUMNS.primary_object_type] == "road",
                _FIAT_COLUMNS.ground_elevation,
            ] = exposure.loc[
                exposure[_FIAT_COLUMNS.primary_object_type] == "road", "elev"
            ]
            del exposure["elev"]
            self.fiat_model.exposure.exposure_db = exposure

        buildings = self.fiat_model.exposure.exposure_geoms[
            self._get_fiat_building_index()
        ].to_crs(dem.spatial_ref.crs_wkt)
        buildings["geometry"] = buildings.geometry.centroid
        x_points = xr.DataArray(buildings["geometry"].x, dims="points")
        y_points = xr.DataArray(buildings["geometry"].y, dims="points")
        buildings["elev"] = (
            dem.sel(x=x_points, y=y_points, band=1, method="nearest").to_numpy()
            * conversion_factor
        )
        exposure = exposure.merge(
            buildings[[_FIAT_COLUMNS.object_id, "elev"]],
            on=_FIAT_COLUMNS.object_id,
            how="left",
        )
        exposure.loc[
            exposure[_FIAT_COLUMNS.primary_object_type] != "road",
            _FIAT_COLUMNS.ground_elevation,
        ] = exposure.loc[exposure[_FIAT_COLUMNS.primary_object_type] != "road", "elev"]
        del exposure["elev"]

    def read_damage_unit(self) -> str:
        if self.fiat_model.exposure.damage_unit is not None:
            return self.fiat_model.exposure.damage_unit
        else:
            self.logger.warning(
                "Delft-FIAT model was missing damage units so '$' was assumed."
            )
            return "$"

    def read_floodmap_type(self) -> FloodmapType:
        # If there is at least on object that uses the area method, use water depths for FA calcs
        if (
            self.fiat_model.exposure.exposure_db[_FIAT_COLUMNS.extraction_method]
            == "area"
        ).any():
            return FloodmapType.water_depth
        else:
            return FloodmapType.water_level

    def create_roads(self) -> Optional[str]:
        # Make sure that FIAT roads are polygons
        if self.config.fiat_roads_name not in self.fiat_model.exposure.geom_names:
            self.logger.warning(
                "Road objects are not available in the FIAT model and thus would not be available in FloodAdapt."
            )
            # TODO check how this naming of output geoms should become more explicit!
            return None

        roads = self.fiat_model.exposure.exposure_geoms[self._get_fiat_road_index()]

        # TODO do we need the lanes column?
        if (
            _FIAT_COLUMNS.segment_length
            not in self.fiat_model.exposure.exposure_db.columns
        ):
            self.logger.warning(
                f"'{_FIAT_COLUMNS.segment_length}' column not present in the FIAT exposure csv. Road impact infometrics cannot be produced."
            )

        # TODO should this should be performed through hydromt-FIAT?
        if not isinstance(roads.geometry.iloc[0], Polygon):
            roads = roads.to_crs(roads.estimate_utm_crs())
            roads.geometry = roads.geometry.buffer(
                self.config.road_width / 2, cap_style=2
            )
            roads = roads.to_crs(self.fiat_model.exposure.crs)
            self.fiat_model.exposure.exposure_geoms[self._get_fiat_road_index()] = roads
            self.logger.info(
                f"FIAT road objects transformed from lines to polygons assuming a road width of {self.config.road_width} meters."
            )

        self._has_roads = True
        return f"{self.config.fiat_roads_name}.gpkg"

    def create_new_developments(self) -> Optional[str]:
        return "new_development_area.gpkg"

    def create_footprints(self) -> Optional[Path]:
        if isinstance(self.config.building_footprints, SpatialJoinModel):
            # Use the provided building footprints
            building_footprints_file = self._check_exists_and_absolute(
                self.config.building_footprints.file
            )

            self.logger.info(
                f"Using building footprints from {Path(building_footprints_file).as_posix()}."
            )
            # Spatially join buildings and map
            # TODO use hydromt method instead
            path = self._join_building_footprints(
                self.config.building_footprints.file,
                self.config.building_footprints.field_name,
            )
            return path
        elif self.config.building_footprints == FootprintsOptions.OSM:
            self.logger.info(
                "Building footprint data will be downloaded from Open Street Maps."
            )
            region = self.fiat_model.region
            if region is None:
                raise ValueError(
                    "No region file found in the FIAT model. Building footprints cannot be created."
                )
            region = region.to_crs(4326)
            if isinstance(region.boundary.to_numpy()[0], MultiLineString):
                polygon = Polygon(
                    region.boundary.to_numpy()[0].envelope
                )  # TODO check if this is correct
            else:
                polygon = Polygon(region.boundary.to_numpy()[0])
            footprints = get_buildings_from_osm(polygon)
            footprints["BF_FID"] = np.arange(1, len(footprints) + 1)
            footprints = footprints[["BF_FID", "geometry"]]
            path = self._join_building_footprints(footprints, "BF_FID")
            return path
        # Then check if geometries are already footprints
        elif isinstance(
            self.fiat_model.exposure.exposure_geoms[
                self._get_fiat_building_index()
            ].geometry.iloc[0],
            (Polygon, MultiPolygon),
        ):
            self.logger.info(
                "Building footprints are already available in the FIAT model geometry files."
            )
            return None
        # check if it is spatially joined and/or exists already
        elif "BF_FID" in self.fiat_model.exposure.exposure_db.columns:
            add_attrs = self.fiat_model.spatial_joins["additional_attributes"]
            fiat_path = Path(self.fiat_model.root)

            if not (add_attrs and "BF_FID" in [attr["name"] for attr in add_attrs]):
                raise KeyError(
                    "While 'BF_FID' column exists, connection to a spatial footprints file is missing."
                )

            ind = [attr["name"] for attr in add_attrs].index("BF_FID")
            footprints = add_attrs[ind]
            footprints_path = fiat_path / footprints["file"]

            if not footprints_path.exists():
                raise FileNotFoundError(
                    f"While 'BF_FID' column exists, building footprints file {footprints_path} not found."
                )

            self.logger.info(
                f"Using the building footprints located at {footprints_path}."
            )
            return footprints_path.relative_to(self.static_path)

        # Other methods
        else:
            self.logger.warning(
                "No building footprints are available. Buildings will be plotted with a default shape in FloodAdapt."
            )
            return None

    def create_bfe(self) -> Optional[BFEModel]:
        if self.config.bfe is None:
            self.logger.warning(
                "No base flood elevation provided. Elevating building relative to base flood elevation will not be possible in FloodAdapt."
            )
            return None

        # TODO can we use hydromt-FIAT?
        bfe_file = self._check_exists_and_absolute(self.config.bfe.file)

        self.logger.info(
            f"Using map from {Path(bfe_file).as_posix()} as base flood elevation."
        )

        # Spatially join buildings and map
        buildings_joined, bfe = self.spatial_join(
            self.fiat_model.exposure.exposure_geoms[self._get_fiat_building_index()],
            bfe_file,
            self.config.bfe.field_name,
        )

        # Make sure in case of multiple values that the max is kept
        buildings_joined = (
            buildings_joined.groupby(_FIAT_COLUMNS.object_id)
            .max(self.config.bfe.field_name)
            .sort_values(by=[_FIAT_COLUMNS.object_id])
            .reset_index()
        )

        # Save the files
        fa_bfe_file = self.static_path / "bfe" / "bfe.gpkg"
        fa_bfe_file.parent.mkdir(parents=True, exist_ok=True)
        bfe.to_file(fa_bfe_file)
        csv_path = fa_bfe_file.parent / "bfe.csv"
        buildings_joined.to_csv(csv_path, index=False)

        # Save attributes
        return BFEModel(
            geom=fa_bfe_file.relative_to(self.static_path).as_posix(),
            table=csv_path.relative_to(self.static_path).as_posix(),
            field_name=self.config.bfe.field_name,
        )

    def create_aggregation_areas(self) -> list[AggregationModel]:
        # TODO split this to 3 methods?
        aggregation_areas = []

        # first check if the FIAT model has existing aggregation areas
        if self.fiat_model.spatial_joins["aggregation_areas"]:
            # Use the aggregation areas from the FIAT model
            for aggr in self.fiat_model.spatial_joins["aggregation_areas"]:
                # Check if the exposure csv has the correct column
                col_name = _FIAT_COLUMNS.aggregation_label.format(name=aggr["name"])
                if col_name not in self.fiat_model.exposure.exposure_db.columns:
                    raise KeyError(
                        f"While aggregation area '{aggr['name']}' exists in the spatial joins of the FIAT model, the column '{col_name}' is missing in the exposure csv."
                    )
                # Check equity config
                if aggr["equity"] is not None:
                    equity_config = EquityModel(
                        census_data=str(
                            self.static_path.joinpath(
                                "templates", "fiat", aggr["equity"]["census_data"]
                            )
                            .relative_to(self.static_path)
                            .as_posix()
                        ),
                        percapitaincome_label=aggr["equity"]["percapitaincome_label"],
                        totalpopulation_label=aggr["equity"]["totalpopulation_label"],
                    )
                else:
                    equity_config = None
                # Make aggregation config
                aggr = AggregationModel(
                    name=aggr["name"],
                    file=str(
                        self.static_path.joinpath("templates", "fiat", aggr["file"])
                        .relative_to(self.static_path)
                        .as_posix()
                    ),
                    field_name=aggr["field_name"],
                    equity=equity_config,
                )
                aggregation_areas.append(aggr)

                self.logger.info(
                    f"Aggregation areas: {aggr.name} from the FIAT model are going to be used."
                )

        # Then check if the user has provided extra aggregation areas in the config
        if self.config.aggregation_areas:
            # Loop through aggr areas given in config
            for aggr in self.config.aggregation_areas:
                # Get name of type of aggregation area
                if aggr.name is not None:
                    aggr_name = aggr.name
                else:
                    aggr_name = Path(aggr.file).stem
                # If aggregation area already in FIAT model raise Error
                if aggr_name in [aggr.name for aggr in aggregation_areas]:
                    raise ValueError(
                        f"Aggregation area '{aggr_name}' already exists in the FIAT model."
                    )
                # Do spatial join of FIAT objects and aggregation areas
                exposure_csv = self.fiat_model.exposure.exposure_db
                buildings_joined, aggr_areas = self.spatial_join(
                    objects=self.fiat_model.exposure.exposure_geoms[
                        self._get_fiat_building_index()
                    ],
                    layer=str(self._check_exists_and_absolute(aggr.file)),
                    field_name=aggr.field_name,
                    rename=_FIAT_COLUMNS.aggregation_label.format(name=aggr_name),
                )
                aggr_path = Path(self.fiat_model.root).joinpath(
                    "exposure", "aggregation_areas", f"{Path(aggr.file).stem}.gpkg"
                )
                aggr_path.parent.mkdir(parents=True, exist_ok=True)
                aggr_areas.to_file(aggr_path)
                exposure_csv = exposure_csv.merge(
                    buildings_joined, on=_FIAT_COLUMNS.object_id, how="left"
                )
                self.fiat_model.exposure.exposure_db = exposure_csv
                # Update spatial joins in FIAT model
                if self.fiat_model.spatial_joins["aggregation_areas"] is None:
                    self.fiat_model.spatial_joins["aggregation_areas"] = []
                self.fiat_model.spatial_joins["aggregation_areas"].append(
                    {
                        "name": aggr_name,
                        "file": aggr_path.relative_to(self.fiat_model.root).as_posix(),
                        "field_name": _FIAT_COLUMNS.aggregation_label.format(
                            name=aggr_name
                        ),
                        "equity": None,  # TODO allow adding equity as well?
                    }
                )
                # Update the aggregation areas list in the config
                aggregation_areas.append(
                    AggregationModel(
                        name=aggr_name,
                        file=aggr_path.relative_to(self.static_path).as_posix(),
                        field_name=_FIAT_COLUMNS.aggregation_label.format(
                            name=aggr_name
                        ),
                    )
                )

        # No config provided, no aggr areas in the model -> try to use the region file as a mock aggregation area
        if (
            not self.fiat_model.spatial_joins["aggregation_areas"]
            and not self.config.aggregation_areas
        ):
            exposure_csv = self.fiat_model.exposure.exposure_db
            region = self.fiat_model.geoms["region"]
            region = region.explode().reset_index()
            region["aggr_id"] = ["region_" + str(i) for i in np.arange(len(region)) + 1]
            aggregation_path = Path(self.fiat_model.root).joinpath(
                "aggregation_areas", "region.geojson"
            )
            if not aggregation_path.parent.exists():
                aggregation_path.parent.mkdir()

            region.to_file(aggregation_path)
            aggr = AggregationModel(
                name="region",
                file=str(aggregation_path.relative_to(self.static_path).as_posix()),
                field_name="aggr_id",
            )
            aggregation_areas.append(aggr)

            # Add column in FIAT
            buildings_joined, _ = self.spatial_join(
                objects=self.fiat_model.exposure.exposure_geoms[
                    self._get_fiat_building_index()
                ],
                layer=region,
                field_name="aggr_id",
                rename=_FIAT_COLUMNS.aggregation_label.format(name="region"),
            )
            exposure_csv = exposure_csv.merge(
                buildings_joined, on=_FIAT_COLUMNS.object_id, how="left"
            )
            self.fiat_model.exposure.exposure_db = exposure_csv
            self.logger.warning(
                "No aggregation areas were available in the FIAT model and none were provided in the config file. The region file will be used as a mock aggregation area."
            )
        return aggregation_areas

    def create_svi(self) -> Optional[SVIModel]:
        if self.config.svi:
            svi_file = self._check_exists_and_absolute(self.config.svi.file)
            exposure_csv = self.fiat_model.exposure.exposure_db
            buildings_joined, svi = self.spatial_join(
                self.fiat_model.exposure.exposure_geoms[
                    self._get_fiat_building_index()
                ],
                svi_file,
                self.config.svi.field_name,
                rename="SVI",
                filter=True,
            )
            # Add column to exposure
            if "SVI" in exposure_csv.columns:
                self.logger.info(
                    f"'SVI' column in the FIAT exposure csv will be replaced by {svi_file.as_posix()}."
                )
                del exposure_csv["SVI"]
            else:
                self.logger.info(
                    f"'SVI' column in the FIAT exposure csv will be filled by {svi_file.as_posix()}."
                )
            exposure_csv = exposure_csv.merge(
                buildings_joined, on=_FIAT_COLUMNS.object_id, how="left"
            )
            self.fiat_model.exposure.exposure_db = exposure_csv

            # Save the spatial file for future use
            svi_path = self.static_path / "templates" / "fiat" / "svi" / "svi.gpkg"
            svi_path.parent.mkdir(parents=True, exist_ok=True)
            svi.to_file(svi_path)
            self.logger.info(
                f"An SVI map can be shown in FloodAdapt GUI using '{self.config.svi.field_name}' column from {svi_file.as_posix()}"
            )

            return SVIModel(
                geom=Path(svi_path.relative_to(self.static_path)).as_posix(),
                field_name="SVI",
            )
        elif "SVI" in self.fiat_model.exposure.exposure_db.columns:
            self.logger.info(
                "'SVI' column present in the FIAT exposure csv. Vulnerability type infometrics can be produced."
            )
            add_attrs = self.fiat_model.spatial_joins["additional_attributes"]
            if "SVI" not in [attr["name"] for attr in add_attrs]:
                self.logger.warning(
                    "No SVI map found to display in the FloodAdapt GUI!"
                )

            ind = [attr["name"] for attr in add_attrs].index("SVI")
            svi = add_attrs[ind]
            svi_path = self.static_path / "templates" / "fiat" / svi["file"]
            self.logger.info(
                f"An SVI map can be shown in FloodAdapt GUI using '{svi['field_name']}' column from {svi['file']}"
            )
            # Save site attributes
            return SVIModel(
                geom=Path(svi_path.relative_to(self.static_path)).as_posix(),
                field_name=svi["field_name"],
            )

        else:
            self.logger.warning(
                "'SVI' column not present in the FIAT exposure csv. Vulnerability type infometrics cannot be produced."
            )
            return None

    ### SFINCS ###
    def create_sfincs_config(self) -> SfincsModel:
        # call these functions before others to make sure water level references are updated
        config = self.create_sfincs_model_config()
        tide_gauge = self.create_tide_gauge()

        sfincs = SfincsModel(
            config=config,
            water_level=self.water_level_references,
            slr_scenarios=self.create_slr(),
            dem=self.create_dem_model(),
            scs=self.create_scs_model(),
            cyclone_track_database=self.create_cyclone_track_database(),
            tide_gauge=tide_gauge,
            river=self.create_rivers(),
            obs_point=self.create_observation_points(),
        )

        return sfincs

    def create_cyclone_track_database(self) -> Optional[CycloneTrackDatabaseModel]:
        if not self.config.cyclones or not self.config.sfincs_offshore:
            self.logger.warning("No cyclones will be available in the database.")
            return None

        if self.config.cyclone_basin:
            basin = self.config.cyclone_basin
        else:
            basin = "ALL"

        name = f"IBTrACS.{basin.value}.v04r01.nc"
        url = f"https://www.ncei.noaa.gov/data/international-best-track-archive-for-climate-stewardship-ibtracs/v04r01/access/netcdf/{name}"
        self.logger.info(f"Downloading cyclone track database from {url}")
        fn = Path(self.root) / "static" / "cyclone_track_database" / name
        fn.parent.mkdir(parents=True, exist_ok=True)

        try:
            urlretrieve(url, fn)
        except Exception:
            raise RuntimeError(f"Could not retrieve cyclone track database from {url}")

        return CycloneTrackDatabaseModel(file=name)

    def create_scs_model(self) -> Optional[SCSModel]:
        if self.config.scs is None:
            return None
        scs_file = self._check_exists_and_absolute(self.config.scs.file)
        db_scs_file = self.static_path / "scs" / scs_file.name
        db_scs_file.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(scs_file, db_scs_file)

        return SCSModel(file=scs_file.name, type=self.config.scs.type)

    def create_dem_model(self) -> DemModel:
        if self.config.dem:
            subgrid_sfincs = Path(self.config.dem.filename)
        else:
            self.logger.warning(
                "No subgrid depth geotiff file provided in the config file. Using the one from the SFINCS model."
            )
            subgrid_sfincs = (
                Path(self.sfincs_overland_model.root) / "subgrid" / "dep_subgrid.tif"
            )

        dem_file = self._check_exists_and_absolute(subgrid_sfincs)
        fa_subgrid_path = self.static_path / "dem" / dem_file.name
        fa_subgrid_path.parent.mkdir(parents=True, exist_ok=True)

        # Check tiles
        tiles_sfincs = Path(self.sfincs_overland_model.root) / "tiles"
        fa_tiles_path = self.static_path / "dem" / "tiles"
        if tiles_sfincs.exists():
            shutil.move(tiles_sfincs, fa_tiles_path)
            if (fa_tiles_path / "index").exists():
                os.rename(fa_tiles_path / "index", fa_tiles_path / "indices")
            self.logger.info(
                "Tiles were already available in the SFINCS model and will directly be used in FloodAdapt."
            )
        else:
            # Make tiles
            fa_tiles_path.mkdir(parents=True)
            self.sfincs_overland_model.setup_tiles(
                path=fa_tiles_path,
                datasets_dep=[{"elevtn": dem_file}],
                zoom_range=[0, 13],
                fmt="png",
            )
            self.logger.info(
                f"Tiles were created using the {subgrid_sfincs.as_posix()} as the elevation map."
            )

        shutil.copy2(dem_file, fa_subgrid_path)
        self._dem_path = fa_subgrid_path
        return DemModel(
            filename=fa_subgrid_path.name, units=us.UnitTypesLength.meters
        )  # always in meters

    def create_sfincs_model_config(self) -> SfincsConfigModel:
        config = SfincsConfigModel(
            csname=self.sfincs_overland_model.crs.name,
            cstype=Cstype(
                self.sfincs_overland_model.crs.type_name.split(" ")[0].lower()
            ),
            offshore_model=self.create_offshore_model(),
            overland_model=self.create_overland_model(),
            floodmap_units=self.unit_system.default_length_units,
            save_simulation=False,
        )

        return config

    def create_slr(self) -> Optional[SlrScenariosModel]:
        if self.config.slr_scenarios is None:
            return None

        self.config.slr_scenarios.file = str(
            self._check_exists_and_absolute(self.config.slr_scenarios.file)
        )
        slr_path = self.static_path / "slr_scenarios"
        slr_path.mkdir()
        new_file = slr_path / Path(self.config.slr_scenarios.file).name
        shutil.copyfile(self.config.slr_scenarios.file, new_file)

        return SlrScenariosModel(
            file=new_file.relative_to(self.static_path).as_posix(),
            relative_to_year=self.config.slr_scenarios.relative_to_year,
        )

    def create_observation_points(self) -> list[ObsPointModel]:
        if self.config.obs_point is None:
            return []

        self.logger.info("Observation points were provided in the config file.")
        return self.config.obs_point

    def create_rivers(self) -> list[RiverModel]:
        src_file = Path(self.sfincs_overland_model.root) / "sfincs.src"
        if not src_file.exists():
            self.logger.warning("No rivers found in the SFINCS model.")
            return []

        df = pd.read_csv(src_file, delim_whitespace=True, header=None, names=["x", "y"])
        river_locs = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df.x, df.y),
            crs=self.sfincs_overland_model.crs,
        )
        rivers = []
        for idx, row in river_locs.iterrows():
            if "dis" in self.sfincs_overland_model.forcing:
                discharge = (
                    self.sfincs_overland_model.forcing["dis"]
                    .sel(index=idx + 1)
                    .to_numpy()
                    .mean()
                )
            else:
                discharge = 0
                self.logger.warning(
                    f"No river discharge conditions were found in the SFINCS model for river {idx}. A default value of 0 will be used."
                )

            river = RiverModel(
                name=f"river_{idx}",
                x_coordinate=row.x,
                y_coordinate=row.y,
                mean_discharge=us.UnitfulDischarge(
                    value=discharge, units=self.unit_system.default_discharge_units
                ),
            )
            rivers.append(river)

        self.logger.info(
            f"{len(river_locs)} river(s) were identified from the SFINCS model and will be available in FloodAdapt for discharge input."
        )

        return rivers

    def create_tide_gauge(self) -> Optional[TideGauge]:
        if self.config.tide_gauge is None:
            self.logger.warning(
                "Tide gauge information not provided. Historical events will not have an option to use gauged data in FloodAdapt!"
            )
            self.logger.warning(
                "No water level references were found. It is assumed that MSL is equal to the datum used in the SFINCS overland model. You can provide these values with the tide_gauge.ref attribute in the site.toml."
            )
            return None

        if self.config.tide_gauge.source == TideGaugeSource.file:
            if self.config.tide_gauge.file is None:
                raise ValueError(
                    "Tide gauge file needs to be provided when 'file' is selected as the source."
                )
            if self.config.tide_gauge.ref is None:
                self.logger.warning(
                    "Tide gauge reference not provided. MSL is assumed as the reference of the water levels in the file."
                )
                self.config.tide_gauge.ref = "MSL"

            tide_gauge_file = self._check_exists_and_absolute(
                self.config.tide_gauge.file
            )
            db_file_path = Path(self.static_path / "tide_gauges") / tide_gauge_file.name

            db_file_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(self.config.tide_gauge.file, db_file_path)

            rel_db_path = Path(db_file_path.relative_to(self.static_path))
            self.logger.warning(
                f"Tide gauge from file {rel_db_path} assumed to be in {self.unit_system.default_length_units}!"
            )
            tide_gauge = TideGauge(
                reference=self.config.tide_gauge.ref,
                description="Observations from file stored in database",
                source=TideGaugeSource.file,
                file=rel_db_path,
                lon=self.config.tide_gauge.lon,
                lat=self.config.tide_gauge.lat,
                units=self.unit_system.default_length_units,
            )

            return tide_gauge

        elif self.config.tide_gauge.source == TideGaugeSource.noaa_coops:
            if self.config.tide_gauge.ref is not None:
                ref = self.config.tide_gauge.ref
            else:
                ref = "MLLW"  # If reference is not provided use MLLW

            self.water_level_references.reference = (
                ref  # update the water level reference
            )

            if self.config.tide_gauge.id is None:
                station_id = self._get_closest_station()
                self.logger.info(
                    "The closest NOAA tide gauge station to the site will be searched."
                )
            else:
                station_id = self.config.tide_gauge.id
                self.logger.info(
                    f"The NOAA tide gauge station with the provided ID {station_id} will be used."
                )
            station = self._get_station_metadata(station_id=station_id, ref=ref)
            if station is not None:
                # Add tide_gauge information in site toml
                tide_gauge = TideGauge(
                    name=station["name"],
                    description=f"observations from '{self.config.tide_gauge.source}' api",
                    source=self.config.tide_gauge.source,
                    reference=ref,
                    ID=int(station["id"]),
                    lon=station["lon"],
                    lat=station["lat"],
                    units=us.UnitTypesLength.meters,  # the api always asks for SI units right now
                )

                local_datum = DatumModel(
                    name=station["datum_name"],
                    height=us.UnitfulLength(
                        value=station["datum"], units=station["units"]
                    ).transform(self.unit_system.default_length_units),
                )
                self.water_level_references.datums.append(local_datum)

                msl = DatumModel(
                    name="MSL",
                    height=us.UnitfulLength(
                        value=station["msl"], units=station["units"]
                    ).transform(self.unit_system.default_length_units),
                )
                #  Check if MSL is already there and if yes replace it
                existing_msl = next(
                    (
                        datum
                        for datum in self.water_level_references.datums
                        if datum.name == "MSL"
                    ),
                    None,
                )
                if existing_msl:
                    self.water_level_references.datums.remove(existing_msl)
                self.water_level_references.datums.append(msl)

                for name in ["MLLW", "MHHW"]:
                    height = us.UnitfulLength(
                        value=station[name.lower()], units=station["units"]
                    ).transform(self.unit_system.default_length_units)

                    wl_info = DatumModel(
                        name=name,
                        height=height,
                    )
                    self.water_level_references.datums.append(wl_info)
            return tide_gauge
        else:
            self.logger.warning(
                f"Tide gauge source not recognized: {self.config.tide_gauge.source}. Historical events will not have an option to use gauged data in FloodAdapt!"
            )
            return None

    def create_offshore_model(self) -> Optional[FloodModel]:
        if self.sfincs_offshore_model is None:
            return None
        # Connect boundary points of overland to output points of offshore
        fn = Path(self.sfincs_overland_model.root) / "sfincs.bnd"
        bnd = pd.read_csv(fn, sep=" ", lineterminator="\n", header=None)
        bnd = bnd.rename(columns={0: "x", 1: "y"})
        bnd_geo = gpd.GeoDataFrame(
            bnd,
            geometry=gpd.points_from_xy(bnd.x, bnd.y),
            crs=self.sfincs_overland_model.config["epsg"],
        )
        obs_geo = bnd_geo.to_crs(4326)
        obs_geo["x"] = obs_geo.geometry.x
        obs_geo["y"] = obs_geo.geometry.y
        del obs_geo["geometry"]
        obs_geo["name"] = [f"bnd_pt{num:02d}" for num in range(1, len(obs_geo) + 1)]
        fn_off = Path(self.sfincs_offshore_model.root) / "sfincs.obs"
        obs_geo.to_csv(
            fn_off,
            sep="\t",
            index=False,
            header=False,
        )
        self.logger.info(
            "Output points of the offshore SFINCS model were reconfigured to the boundary points of the overland SFINCS model."
        )

        return FloodModel(
            name="offshore",
            reference=self.config.sfincs_offshore.reference,
            vertical_offset=self.config.sfincs_offshore.vertical_offset,
        )

    def create_overland_model(self) -> FloodModel:
        return FloodModel(
            name="overland",
            reference=self.config.sfincs_overland.reference,
        )

    ### SITE ###
    def create_site_config(self) -> Site:
        """Create the site configuration for the FloodAdapt model.

        The order of these functions is important!
        1. Create the SFINCS model.
            needs: water level references
            provides: updated water level references with optional tide gauge
        2. Create the FIAT model.
            needs: water level references and optional probabilistic event set
            provides: svi and exposure geometries
        3. Create the GUI model. (requires water level references and FIAT model to be updated)
            needs: water level references and FIAT model to be updated
            provides: gui model with output layers, visualization layers and plotting.

        """
        sfincs = self.create_sfincs_config()
        self.add_probabilistic_set()
        fiat = self.create_fiat_model()
        gui = self.create_gui_config()

        # Order doesnt matter from here
        lon, lat = self.read_location()
        std_objs = self.set_standard_objects()
        description = (
            self.config.description if self.config.description else self.config.name
        )

        config = Site(
            name=self.config.name,
            description=description,
            lat=lat,
            lon=lon,
            fiat=fiat,
            gui=gui,
            sfincs=sfincs,
            standard_objects=std_objs,
        )
        return config

    def read_location(self) -> tuple[float, float]:
        # Get center of area of interest
        if not self.fiat_model.region.empty:
            center = self.fiat_model.region.dissolve().centroid.to_crs(4326)[0]
        else:
            center = (
                self.fiat_model.exposure.exposure_geoms[self._get_fiat_building_index()]
                .dissolve()
                .centroid.to_crs(4326)[0]
            )
        return center.x, center.y

    def create_gui_config(self) -> GuiModel:
        gui = GuiModel(
            units=self.unit_system,
            plotting=self.create_hazard_plotting_config(),
            output_layers=self.create_output_layers_config(),
            visualization_layers=self.create_visualization_layers(),
        )

        return gui

    def create_default_units(self) -> GuiUnitModel:
        if self.config.unit_system == UnitSystems.imperial:
            return GuiUnitModel.imperial()
        elif self.config.unit_system == UnitSystems.metric:
            return GuiUnitModel.metric()
        else:
            raise ValueError(
                f"Unit system {self.config.unit_system} not recognized. Please choose 'imperial' or 'metric'."
            )

    def create_visualization_layers(self) -> VisualizationLayers:
        visualization_layers = VisualizationLayers()
        if self._svi is not None:
            visualization_layers.add_layer(
                name="svi",
                long_name="Social Vulnerability Index (SVI)",
                path=str(self.static_path / self._svi.geom),
                database_path=self.root,
                field_name="SVI",
                bins=[0.05, 0.2, 0.4, 0.6, 0.8],
            )
        return visualization_layers

    def create_output_layers_config(self) -> OutputLayers:
        # Read default colors from template
        fd_max = self.config.gui.max_flood_depth
        ad_max = self.config.gui.max_aggr_dmg
        ftd_max = self.config.gui.max_footprint_dmg
        b_max = self.config.gui.max_benefits

        benefits_layer = None
        if self.config.probabilistic_set is not None:
            benefits_layer = BenefitsLayer(
                bins=[0, 0.01, 0.02 * b_max, 0.2 * b_max, b_max],
                colors=[
                    "#FF7D7D",
                    "#FFFFFF",
                    "#DCEDC8",
                    "#AED581",
                    "#7CB342",
                    "#33691E",
                ],
                threshold=0.0,
            )

        output_layers = OutputLayers(
            floodmap=FloodMapLayer(
                bins=[0.2 * fd_max, 0.6 * fd_max, fd_max],
                colors=["#D7ECFB", "#8ABDDD", "#1C73A4", "#081D58"],
                zbmax=-9999,
                depth_min=0.0,
            ),
            aggregation_dmg=AggregationDmgLayer(
                bins=[0.00001, 0.1 * ad_max, 0.25 * ad_max, 0.5 * ad_max, ad_max],
                colors=[
                    "#FFFFFF",
                    "#FEE9CE",
                    "#FDBB84",
                    "#FC844E",
                    "#E03720",
                    "#860000",
                ],
            ),
            footprints_dmg=FootprintsDmgLayer(
                bins=[0.00001, 0.06 * ftd_max, 0.2 * ftd_max, 0.4 * ftd_max, ftd_max],
                colors=[
                    "#FFFFFF",
                    "#FEE9CE",
                    "#FDBB84",
                    "#FC844E",
                    "#E03720",
                    "#860000",
                ],
            ),
            benefits=benefits_layer,
        )
        return output_layers

    def create_hazard_plotting_config(self) -> PlottingModel:
        datum_names = [datum.name for datum in self.water_level_references.datums]
        if "MHHW" in datum_names:
            amplitude = (
                self.water_level_references.get_datum("MHHW").height
                - self.water_level_references.get_datum("MSL").height
            )
            self.logger.info(
                f"The default tidal amplitude in the GUI will be {amplitude.transform(self.unit_system.default_length_units)}, calculated as the difference between MHHW and MSL from the tide gauge data."
            )
        else:
            amplitude = us.UnitfulLength(
                value=0.0, units=self.unit_system.default_length_units
            )
            self.logger.warning(
                "The default tidal amplitude in the GUI will be 0.0, since no tide-gauge water levels are available. You can change this in the site.toml with the 'gui.tide_harmonic_amplitude' attribute."
            )

        ref = "MSL"
        if ref not in datum_names:
            self.logger.warning(
                f"The Mean Sea Level (MSL) datum is not available in the site.toml. The synthetic tide will be created relative to the main reference: {self.water_level_references.reference}."
            )
            ref = self.water_level_references.reference

        plotting = PlottingModel(
            synthetic_tide=SyntheticTideModel(
                harmonic_amplitude=amplitude,
                datum=ref,
            ),
            excluded_datums=self.config.excluded_datums,
        )

        return plotting

    def create_infometrics(self):
        """
        Copy the infometrics and infographics templates to the appropriate location and modifies the metrics_config.toml files.

        This method copies the templates from the 'infometrics' and 'infographics' folders to the 'static/templates' folder in the root directory.
        It then modifies the 'metrics_config.toml' and 'metrics_config_risk.toml' files by updating the 'aggregateBy' attribute with the names
        of the aggregations defined in the 'fiat' section of the 'site_attrs' attribute.
        """
        # TODO there should be generalized infometric queries with NSI or OSM, and with SVI or without. Then Based on the user input these should be chosen automatically
        templates_path = Path(__file__).parent.resolve().joinpath("templates")

        # Create template folder
        path_im = self.root.joinpath("static", "templates", "infometrics")
        path_im.mkdir()

        # Copy mandatory metric configs
        path_im_temp = templates_path.joinpath("infometrics")
        for file in path_im_temp.glob("*.toml"):
            shutil.copy2(file, path_im)

        self._create_optional_infometrics(templates_path, path_im)

        files = list(path_im.glob("*metrics_config*.toml"))
        # Update aggregation areas in metrics config
        for file in files:
            file = path_im.joinpath(file)
            with open(file, "rb") as f:
                attrs = tomli.load(f)

            # add aggration levels
            if self._aggregation_areas is None:
                self._aggregation_areas = self.create_aggregation_areas()
            attrs["aggregateBy"] = [aggr.name for aggr in self._aggregation_areas]

            # take out road metrics if needed
            if not self._has_roads:
                attrs["queries"] = [
                    query
                    for query in attrs["queries"]
                    if "road" not in query["name"].lower()
                ]

            # Replace Damage Unit
            # TODO do this in a better manner
            for i, query in enumerate(attrs["queries"]):
                if "$" in query["long_name"]:
                    query["long_name"] = query["long_name"].replace(
                        "$", self.read_damage_unit()
                    )

            # replace the SVI threshold if needed
            if self.config.svi:
                for i, query in enumerate(attrs["queries"]):
                    query["filter"] = query["filter"].replace(
                        "SVI_threshold", str(self.config.svi.threshold)
                    )

            with open(file, "wb") as f:
                tomli_w.dump(attrs, f)

    def _create_optional_infometrics(self, templates_path: Path, path_im: Path):
        # If infographics are going to be created in FA, get template metric configurations
        if not self.config.infographics:
            return

        # Check what type of infographics should be used
        if self.config.unit_system == UnitSystems.imperial:
            metrics_folder_name = "US_NSI"
            self.logger.info(
                "Default NSI infometrics and infographics will be created."
            )
        elif self.config.unit_system == UnitSystems.metric:
            metrics_folder_name = "OSM"
            self.logger.info(
                "Default OSM infometrics and infographics will be created."
            )
        else:
            raise ValueError(
                f"Unit system {self.config.unit_system} is not recognized. Please choose 'imperial' or 'metric'."
            )

        if self.config.svi is not None:
            svi_folder_name = "with_SVI"
        else:
            svi_folder_name = "without_SVI"

        # Copy metrics config for infographics
        path_0 = templates_path.joinpath(
            "infometrics", metrics_folder_name, svi_folder_name
        )
        for file in path_0.glob("*.toml"):
            shutil.copy2(file, path_im)

        # Copy additional risk config
        file = templates_path.joinpath(
            "infometrics",
            metrics_folder_name,
            "metrics_additional_risk_configs.toml",
        )
        shutil.copy2(file, path_im)

        # Copy infographics config
        path_ig_temp = templates_path.joinpath("infographics", metrics_folder_name)
        path_ig = self.root.joinpath("static", "templates", "infographics")
        path_ig.mkdir()
        files_ig = ["styles.css", "config_charts.toml"]

        if self.config.svi is not None:
            files_ig.append("config_risk_charts.toml")
            files_ig.append("config_people.toml")

        if self._has_roads:
            files_ig.append("config_roads.toml")

        for file in files_ig:
            shutil.copy2(path_ig_temp.joinpath(file), path_ig.joinpath(file))

        # Copy images
        path_0 = templates_path.joinpath("infographics", "images")
        path_1 = self.root.joinpath("static", "templates", "infographics", "images")
        shutil.copytree(path_0, path_1)

    def add_static_files(self):
        """
        Copy static files from the 'templates' folder to the 'static' folder.

        This method iterates over a list of folders and copies the contents of each folder from the 'templates' directory
        to the corresponding folder in the 'static' directory.
        """
        templates_path = Path(__file__).parent.resolve().joinpath("templates")
        folders = ["icons", "green_infra_table"]
        for folder in folders:
            path_0 = templates_path.joinpath(folder)
            path_1 = self.static_path / folder
            shutil.copytree(path_0, path_1)

    def add_probabilistic_set(self):
        # Copy prob set if given
        if self.config.probabilistic_set:
            self.logger.info(
                f"Probabilistic event set imported from {self.config.probabilistic_set}"
            )
            prob_event_name = Path(self.config.probabilistic_set).name
            path_db = self.root.joinpath("input", "events", prob_event_name)
            shutil.copytree(self.config.probabilistic_set, path_db)
            self._probabilistic_set_name = prob_event_name
        else:
            self.logger.warning(
                "Probabilistic event set not provided. Risk scenarios cannot be run in FloodAdapt."
            )
            self._probabilistic_set_name = None

    ### HELPER FUNCTIONS ###
    def make_folder_structure(self) -> None:
        """
        Create the folder structure for the database.

        This method creates the necessary folder structure for the FloodAdapt database, including
        the input and static folders. It also creates subfolders within the input and
        static folders based on a predefined list of names.
        """
        self.logger.info("Preparing the database folder structure.")
        inputs = [
            "events",
            "projections",
            "measures",
            "strategies",
            "scenarios",
            "benefits",
        ]
        for name in inputs:
            (self.root / "input" / name).mkdir(parents=True, exist_ok=True)

        # Prepare static folder structure
        folders = ["templates", "config"]
        for name in folders:
            (self.static_path / name).mkdir(parents=True, exist_ok=True)

    def _check_exists_and_absolute(self, path: str) -> Path:
        """Check if the path is absolute or relative and return a Path object. Raises an error if the path is not valid."""
        if not Path(path).exists():
            raise FileNotFoundError(f"Path {path} does not exist.")

        if Path(path).is_absolute():
            return Path(path)
        else:
            raise ValueError(f"Path {path} is not absolute.")

    def _join_building_footprints(
        self, building_footprints: gpd.GeoDataFrame, field_name: str
    ) -> Path:
        """
        Join building footprints with existing building data and updates the exposure CSV.

        Args:
            building_footprints (GeoDataFrame): GeoDataFrame containing the building footprints to be joined.
            field_name (str): The field name to use for the spatial join.

        Returns
        -------
        This method performs the following steps:
        1. Reads the exposure CSV file.
        2. Performs a spatial join between the buildings and building footprints.
        3. Ensures that in case of multiple values, the first is kept.
        4. Creates a folder to store the building footprints.
        5. Saves the spatial file for future use.
        6. Merges the joined buildings with the exposure CSV and saves it.
        7. Updates the site attributes with the relative path to the saved building footprints.
        8. Logs the location where the building footprints are saved.
        """
        buildings = self.fiat_model.exposure.exposure_geoms[
            self._get_fiat_building_index()
        ]
        exposure_csv = self.fiat_model.exposure.exposure_db
        if "BF_FID" in exposure_csv.columns:
            self.logger.warning(
                "Column 'BF_FID' already exists in the exposure columns and will be replaced."
            )
            del exposure_csv["BF_FID"]
        buildings_joined, building_footprints = self.spatial_join(
            buildings,
            building_footprints,
            field_name,
            rename="BF_FID",
            filter=True,
        )
        # Make sure in case of multiple values that the first is kept
        buildings_joined = (
            buildings_joined.groupby(_FIAT_COLUMNS.object_id)
            .first()
            .sort_values(by=[_FIAT_COLUMNS.object_id])
        )
        # Create folder
        bf_folder = Path(self.fiat_model.root) / "exposure" / "building_footprints"
        bf_folder.mkdir(parents=True, exist_ok=True)

        # Save the spatial file for future use
        geo_path = bf_folder / "building_footprints.gpkg"
        building_footprints.to_file(geo_path)

        # Save to exposure csv
        exposure_csv = exposure_csv.merge(
            buildings_joined, on=_FIAT_COLUMNS.object_id, how="left"
        )

        # Set model building footprints
        self.fiat_model.building_footprint = building_footprints
        self.fiat_model.exposure.exposure_db = exposure_csv

        # Save site attributes
        buildings_path = geo_path.relative_to(self.static_path)
        self.logger.info(
            f"Building footprints saved at {(self.static_path / buildings_path).resolve().as_posix()}"
        )

        return buildings_path

    def _clip_hazard_extend(self, clip_footprints=True):
        """
        Clip the exposure data to the bounding box of the hazard data.

        This method clips the exposure data to the bounding box of the hazard data. It creates a GeoDataFrame
        from the hazard polygons, and then uses the `gpd.clip` function to clip the exposure geometries to the
        bounding box of the hazard polygons. If the exposure data contains roads, it is split into two separate
        GeoDataFrames: one for buildings and one for roads. The clipped exposure data is then saved back to the
        `exposure_db` attribute of the `FiatModel` object.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        gdf = self.fiat_model.exposure.get_full_gdf(
            self.fiat_model.exposure.exposure_db
        )
        crs = gdf.crs
        sfincs_extend = self.sfincs_overland_model.region
        sfincs_extend = sfincs_extend.to_crs(crs)

        # Clip the fiat region
        clipped_region = self.fiat_model.region.to_crs(crs).clip(sfincs_extend)
        self.fiat_model.geoms["region"] = clipped_region

        # Clip the exposure geometries
        # Filter buildings and roads
        road_inds = gdf[_FIAT_COLUMNS.primary_object_type].str.contains("road")
        # Ensure road_inds is a boolean Series
        if not road_inds.dtype == bool:
            road_inds = road_inds.astype(bool)
        # Clip buildings
        gdf_buildings = gdf[~road_inds]
        gdf_buildings = self._clip_gdf(
            gdf_buildings, clipped_region, predicate="within"
        ).reset_index(drop=True)

        if road_inds.any():
            # Clip roads
            gdf_roads = gdf[road_inds]
            gdf_roads = self._clip_gdf(
                gdf_roads, clipped_region, predicate="within"
            ).reset_index(drop=True)

            idx_buildings = self.fiat_model.exposure.geom_names.index(
                self.config.fiat_buildings_name
            )
            idx_roads = self.fiat_model.exposure.geom_names.index(
                self.config.fiat_roads_name
            )
            self.fiat_model.exposure.exposure_geoms[idx_buildings] = gdf_buildings[
                [_FIAT_COLUMNS.object_id, "geometry"]
            ]
            self.fiat_model.exposure.exposure_geoms[idx_roads] = gdf_roads[
                [_FIAT_COLUMNS.object_id, "geometry"]
            ]
            gdf = pd.concat([gdf_buildings, gdf_roads])
        else:
            gdf = gdf_buildings
            self.fiat_model.exposure.exposure_geoms[0] = gdf[
                [_FIAT_COLUMNS.object_id, "geometry"]
            ]

        # Save exposure dataframe
        del gdf["geometry"]
        self.fiat_model.exposure.exposure_db = gdf.reset_index(drop=True)

        # Clip the building footprints
        fieldname = "BF_FID"
        if clip_footprints and not self.fiat_model.building_footprint.empty:
            # Get buildings after filtering and their footprint id
            self.fiat_model.building_footprint = self.fiat_model.building_footprint[
                self.fiat_model.building_footprint[fieldname].isin(
                    gdf_buildings[fieldname]
                )
            ].reset_index(drop=True)

    @staticmethod
    def _clip_gdf(
        gdf1: gpd.GeoDataFrame, gdf2: gpd.GeoDataFrame, predicate: str = "within"
    ):
        gdf_new = gpd.sjoin(gdf1, gdf2, how="inner", predicate=predicate)
        gdf_new = gdf_new.drop(
            columns=[
                col
                for col in gdf_new.columns
                if col.endswith("_right") or (col in gdf2.columns and col != "geometry")
            ]
        )

        return gdf_new

    @staticmethod
    def spatial_join(
        objects: gpd.GeoDataFrame,
        layer: Union[str, gpd.GeoDataFrame],
        field_name: str,
        rename: Optional[str] = None,
        filter: Optional[bool] = False,
    ) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """
        Perform a spatial join between two GeoDataFrames.

        Args:
            objects (gpd.GeoDataFrame): The GeoDataFrame representing the objects.
            layer (Union[str, gpd.GeoDataFrame]): The GeoDataFrame or file path of the layer to join with.
            field_name (str): The name of the field to use for the join.
            rename (Optional[str], optional): The new name to assign to the joined field. Defaults to None.

        Returns
        -------
            tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]: A tuple containing the joined GeoDataFrame and the layer GeoDataFrame.

        """
        # Read in layer and keep only column of interest
        if not isinstance(layer, gpd.GeoDataFrame):
            layer = gpd.read_file(layer)
        layer = layer[[field_name, "geometry"]]
        layer = layer.to_crs(objects.crs)
        if field_name in objects.columns:
            layer = layer.rename(columns={field_name: "layer_field"})
            field_name = "layer_field"
        # Spatial join of the layers
        objects_joined = objects.sjoin(layer, how="left", predicate="intersects")

        # Keep only the first intersection for each object
        objects_joined = (
            objects_joined.groupby(_FIAT_COLUMNS.object_id).first().reset_index()
        )

        # if needed filter out unused objects in the layer
        if filter:
            layer_inds = objects_joined["index_right"].dropna().unique()
            layer = layer.iloc[np.sort(layer_inds)].reset_index(drop=True)
        objects_joined = objects_joined[[_FIAT_COLUMNS.object_id, field_name]]
        # rename field if provided
        if rename:
            objects_joined = objects_joined.rename(columns={field_name: rename})
            layer = layer.rename(columns={field_name: rename})
        return objects_joined, layer

    def _get_fiat_building_index(self) -> int:
        return self.fiat_model.exposure.geom_names.index(
            self.config.fiat_buildings_name
        )

    def _get_fiat_road_index(self) -> int:
        return self.fiat_model.exposure.geom_names.index(self.config.fiat_roads_name)

    def _get_closest_station(self):
        # Get available stations from source
        obs_data = obs.source(self.config.tide_gauge.source)
        obs_data.get_active_stations()
        obs_stations = obs_data.gdf()
        # Calculate distance from SFINCS region to all available stations in degrees
        obs_stations["distance"] = obs_stations.distance(
            self.sfincs_overland_model.region.to_crs(4326).geometry.item()
        )
        # Get the closest station and its distance in meters
        closest_station = obs_stations[
            obs_stations["distance"] == obs_stations["distance"].min()
        ]
        distance = round(
            closest_station.to_crs(self.sfincs_overland_model.region.crs)
            .distance(self.sfincs_overland_model.region.geometry.item())
            .item(),
            0,
        )

        distance = us.UnitfulLength(value=distance, units=us.UnitTypesLength.meters)
        self.logger.info(
            f"The closest tide gauge from {self.config.tide_gauge.source} is located {distance.transform(self.unit_system.default_length_units)} from the SFINCS domain"
        )
        # Check if user provided max distance
        # TODO make sure units are explicit for max_distance
        if self.config.tide_gauge.max_distance is not None:
            units_new = self.config.tide_gauge.max_distance.units
            distance_new = us.UnitfulLength(
                value=distance.convert(units_new), units=units_new
            )
            if distance_new.value > self.config.tide_gauge.max_distance.value:
                self.logger.warning(
                    f"This distance is larger than the 'max_distance' value of {self.config.tide_gauge.max_distance.value} {units_new} provided in the config file. The station cannot be used."
                )
                return None

        # get station id
        station_id = closest_station["id"].item()

        return station_id

    def _get_station_metadata(self, station_id: str, ref: str = "MLLW"):
        """
        Find the closest tide gauge station to the SFINCS domain and retrieves its metadata.

        Args:
            ref (str, optional): The reference level for water level measurements. Defaults to "MLLW".

        Returns
        -------
            dict: A dictionary containing the metadata of the closest tide gauge station.
                The dictionary includes the following keys:
                - "id": The station ID.
                - "name": The station name.
                - "datum": The difference between the station's datum and the reference level.
                - "datum_name": The name of the datum used by the station.
                - "msl": The difference between the Mean Sea Level (MSL) and the reference level.
                - "reference": The reference level used for water level measurements.
                - "lon": The longitude of the station.
                - "lat": The latitude of the station.
        """
        # Get available stations from source
        obs_data = obs.source(self.config.tide_gauge.source)
        # read station metadata
        station_metadata = obs_data.get_meta_data(station_id)
        # TODO check if all stations can be used? Tidal attr?
        # Get water levels by using the ref provided
        datum_name = station_metadata["datums"]["OrthometricDatum"]
        datums = station_metadata["datums"]["datums"]
        names = [datum["name"] for datum in datums]

        ref_value = datums[names.index(ref)]["value"]

        meta = {
            "id": station_id,
            "name": station_metadata["name"],
            "datum": round(datums[names.index(datum_name)]["value"] - ref_value, 3),
            "datum_name": datum_name,
            "msl": round(datums[names.index("MSL")]["value"] - ref_value, 3),
            "mllw": round(datums[names.index("MLLW")]["value"] - ref_value, 3),
            "mhhw": round(datums[names.index("MHHW")]["value"] - ref_value, 3),
            "reference": ref,
            "units": station_metadata["datums"]["units"],
            "lon": station_metadata["lng"],
            "lat": station_metadata["lat"],
        }

        self.logger.info(
            f"The tide gauge station '{station_metadata['name']}' from {self.config.tide_gauge.source} will be used to download nearshore historical water level time-series."
        )

        self.logger.info(
            f"The station metadata will be used to fill in the water_level attribute in the site.toml. The reference level will be {ref}."
        )

        return meta

    def _get_bin_colors(self):
        """
        Retrieve the bin colors from the bin_colors.toml file.

        Returns
        -------
            dict: A dictionary containing the bin colors.
        """
        templates_path = Path(__file__).parent.resolve().joinpath("templates")
        with open(
            templates_path.joinpath("output_layers", "bin_colors.toml"), "rb"
        ) as f:
            bin_colors = tomli.load(f)
        return bin_colors


if __name__ == "__main__":
    while True:
        config_path = Path(
            input(
                "Please provide the path to the database creation configuration toml: \n"
            )
        )
        try:
            config = ConfigModel.read(config_path)
            dbs = DatabaseBuilder(config)
            dbs.build()
        except Exception as e:
            print(e)
        quit = input("Do you want to quit? (y/n)")
        if quit == "y":
            exit()



import logging
import os
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any, ClassVar, List, Type

import tomli
from pydantic import BaseModel, field_serializer

from flood_adapt.config.sfincs import RiverModel
from flood_adapt.misc.log import FloodAdaptLogging


### ENUMS ###
class ForcingType(str, Enum):
    """Enum class for the different types of forcing parameters.

    Attributes
    ----------
    RAINFALL : The type of forcing parameter for rainfall.
    WIND : The type of forcing parameter for wind.
    DISCHARGE : The type of forcing parameter for discharge.
    WATERLEVEL : The type of forcing parameter for water level.
    """

    WIND = "WIND"
    RAINFALL = "RAINFALL"
    DISCHARGE = "DISCHARGE"
    WATERLEVEL = "WATERLEVEL"


class ForcingSource(str, Enum):
    """Enum class for the different sources of forcing parameters."""

    MODEL = "MODEL"  # 'our' hindcast/ sfincs offshore model
    TRACK = "TRACK"  # 'our' hindcast/ sfincs offshore model + (shifted) hurricane
    CSV = "CSV"  # user provided csv file
    NETCDF = "NETCDF"  # user provided netcdf file

    SYNTHETIC = "SYNTHETIC"  # synthetic data
    CONSTANT = "CONSTANT"  # synthetic data

    GAUGED = "GAUGED"  # data downloaded from a gauge
    METEO = "METEO"  # external hindcast data

    NONE = "NONE"  # no forcing data


class IForcing(BaseModel, ABC):
    """BaseModel describing the expected variables and data types for forcing parameters of hazard model."""

    class Config:
        arbitrary_types_allowed = True

    type: ForcingType
    source: ForcingSource
    logger: ClassVar[logging.Logger] = FloodAdaptLogging.getLogger("Forcing")

    @classmethod
    def load_file(cls, path: Path):
        with open(path, mode="rb") as fp:
            toml_data = tomli.load(fp)
        return cls.load_dict(toml_data)

    @classmethod
    def load_dict(cls, attrs):
        return cls.model_validate(attrs)

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Override the default model_dump to include class variables `type` and `source`."""
        data = super().model_dump(**kwargs)
        data.update({"type": self.type, "source": self.source})
        return data

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        """Save additional data of the forcing."""
        return

    @field_serializer("path", check_fields=False)
    @classmethod
    def serialize_path(cls, value: Path) -> str:
        """Serialize filepath-like fields by saving only the filename. It is assumed that the file will be saved in the same directory."""
        return value.name


class IDischarge(IForcing):
    type: ForcingType = ForcingType.DISCHARGE
    river: RiverModel


class IRainfall(IForcing):
    type: ForcingType = ForcingType.RAINFALL


class IWind(IForcing):
    type: ForcingType = ForcingType.WIND


class IWaterlevel(IForcing):
    type: ForcingType = ForcingType.WATERLEVEL


class IForcingFactory:
    @classmethod
    @abstractmethod
    def load_file(cls, toml_file: Path) -> IForcing:
        """Create a forcing object from a TOML file."""
        ...

    @classmethod
    @abstractmethod
    def load_dict(cls, attrs: dict[str, Any] | IForcing) -> IForcing:
        """Create a forcing object from a dictionary of attributes."""
        ...

    @classmethod
    @abstractmethod
    def read_forcing(
        cls,
        filepath: Path,
    ) -> tuple[Type[IForcing], ForcingType, ForcingSource]:
        """Extract forcing class, type and source from a TOML file."""
        ...

    @classmethod
    @abstractmethod
    def get_forcing_class(
        cls, type: ForcingType, source: ForcingSource
    ) -> Type[IForcing]:
        """Get the forcing class corresponding to the type and source."""
        ...

    @classmethod
    @abstractmethod
    def list_forcing_types(cls) -> List[ForcingType]:
        """List all available forcing types."""
        ...

    @classmethod
    @abstractmethod
    def list_forcing_classes(cls) -> List[Type[IForcing]]:
        """List all available forcing classes."""
        ...

    @classmethod
    @abstractmethod
    def list_forcing_types_and_sources(cls) -> List[tuple[ForcingType, ForcingSource]]:
        """List all available combinations of forcing types and sources."""
        ...



import os
from pathlib import Path
from typing import Optional

from pydantic import BaseModel, Field

from flood_adapt.misc.utils import resolve_filepath, save_file_to_database
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.object_model import Object


class PhysicalProjection(BaseModel):
    """The accepted input for a physical projection in FloodAdapt.

    Attributes
    ----------
    sea_level_rise : us.UnitfulLength
        The sea level rise in meters. Default=us.UnitfulLength(0.0, us.UnitTypesLength.meters).
    subsidence : us.UnitfulLength
        The subsidence in meters. Default=us.UnitfulLength(0.0, us.UnitTypesLength.meters).
    rainfall_multiplier : float
        The rainfall multiplier. Default = 1.0.
    storm_frequency_increase : float
        The storm frequency increase as a percentage. Default = 0.0.
    """

    sea_level_rise: us.UnitfulLength = us.UnitfulLength(
        value=0.0, units=us.UnitTypesLength.meters
    )
    subsidence: us.UnitfulLength = us.UnitfulLength(
        value=0.0, units=us.UnitTypesLength.meters
    )
    rainfall_multiplier: float = Field(default=1.0, ge=0.0)
    storm_frequency_increase: float = 0.0


class SocioEconomicChange(BaseModel):
    """The accepted input for socio-economic change in FloodAdapt.

    Attributes
    ----------
    population_growth_existing : float
        The existing population growth rate. default=0.0
    economic_growth : float
        The economic growth rate. default=0.0.
    population_growth_new : float
        The population growth rate for new developments. default=0.0.
    new_development_elevation : Optional[us.UnitfulLengthRefValue]
        The elevation of new developments. default=None.
    new_development_shapefile : Optional[str]
        The path to the shapefile of new developments. default=None.
    """

    population_growth_existing: Optional[float] = 0.0
    economic_growth: Optional[float] = 0.0

    population_growth_new: Optional[float] = 0.0
    new_development_elevation: Optional[us.UnitfulLengthRefValue] = None
    new_development_shapefile: Optional[str] = None


class Projection(Object):
    """The accepted input for a projection in FloodAdapt.

    A projection is a combination of a physical projection and a socio-economic change.

    Attributes
    ----------
    physical_projection : PhysicalProjection
        The physical projection model. Contains information about hazard drivers.
    socio_economic_change : SocioEconomicChange
        The socio-economic change model. Contains information about impact drivers.

    """

    physical_projection: PhysicalProjection
    socio_economic_change: SocioEconomicChange

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        if self.socio_economic_change.new_development_shapefile:
            src_path = resolve_filepath(
                "projections",
                self.name,
                self.socio_economic_change.new_development_shapefile,
            )
            path = save_file_to_database(src_path, Path(output_dir))

            # Update the shapefile path in the object so it is saved in the toml file as well
            self.socio_economic_change.new_development_shapefile = path.name



from pathlib import Path

from flood_adapt.database_builder.database_builder import ConfigModel, DatabaseBuilder


def create_database(config_path: Path) -> None:
    """Create a new database from a configuration file or dictionary.

    Parameters
    ----------
    config : Path
        The path to the configuration file
    """
    config = ConfigModel.read(config_path)

    DatabaseBuilder(config=config).build()



import gc
import os
import shutil
import time
from pathlib import Path
from typing import Any, Literal, Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from cht_cyclones.tropical_cyclone import TropicalCyclone
from geopandas import GeoDataFrame
from plotly.express import line
from plotly.express.colors import sample_colorscale

from flood_adapt.config.config import Settings
from flood_adapt.config.site import Site
from flood_adapt.dbs_classes.dbs_benefit import DbsBenefit
from flood_adapt.dbs_classes.dbs_event import DbsEvent
from flood_adapt.dbs_classes.dbs_measure import DbsMeasure
from flood_adapt.dbs_classes.dbs_projection import DbsProjection
from flood_adapt.dbs_classes.dbs_scenario import DbsScenario
from flood_adapt.dbs_classes.dbs_static import DbsStatic
from flood_adapt.dbs_classes.dbs_strategy import DbsStrategy
from flood_adapt.dbs_classes.interface.database import IDatabase
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.path_builder import (
    TopLevelDir,
    db_path,
)
from flood_adapt.misc.utils import finished_file_exists
from flood_adapt.objects.benefits.benefits import Benefit
from flood_adapt.objects.events.events import Event
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.scenarios.scenarios import Scenario
from flood_adapt.workflows.benefit_runner import BenefitRunner
from flood_adapt.workflows.scenario_runner import ScenarioRunner


class Database(IDatabase):
    """Implementation of IDatabase class that holds the site information and has methods to get static data info, and all the input information.

    Additionally it can manipulate (add, edit, copy and delete) any of the objects in the input.
    """

    _instance = None

    database_path: Union[str, os.PathLike]
    database_name: str
    _init_done: bool = False

    base_path: Path
    input_path: Path
    static_path: Path
    output_path: Path

    _site: Site

    _events: DbsEvent
    _scenarios: DbsScenario
    _strategies: DbsStrategy
    _measures: DbsMeasure
    _projections: DbsProjection
    _benefits: DbsBenefit

    _static: DbsStatic

    def __new__(cls, *args, **kwargs):
        if not cls._instance:  # Singleton pattern
            cls._instance = super(Database, cls).__new__(cls)
        return cls._instance

    def __init__(
        self,
        database_path: Union[str, os.PathLike, None] = None,
        database_name: Optional[str] = None,
    ) -> None:
        """
        Initialize the DatabaseController object.

        Parameters
        ----------
        database_path : Union[str, os.PathLike]
            The path to the database root
        database_name : str
            The name of the database.
        -----
        """
        if database_path is None or database_name is None:
            if not self._init_done:
                raise ValueError(
                    """Database path and name must be provided for the first initialization.
                    To do this, run `flood_adapt.api.static.read_database(database_path, site_name)` first."""
                )
            else:
                return  # Skip re-initialization

        if (
            self._init_done
            and self.database_path == database_path
            and self.database_name == database_name
        ):
            return  # Skip re-initialization

        # If the database is not initialized, or a new path or name is provided, (re-)initialize
        re_option = "re-" if self._init_done else ""
        self.logger = FloodAdaptLogging.getLogger("Database")
        self.logger.info(
            f"{re_option}initializing database to {database_name} at {database_path}".capitalize()
        )
        self.database_path = database_path
        self.database_name = database_name

        # Set the paths

        self.base_path = Path(database_path) / database_name
        self.input_path = db_path(TopLevelDir.input)
        self.static_path = db_path(TopLevelDir.static)
        self.output_path = db_path(TopLevelDir.output)

        self._site = Site.load_file(self.static_path / "config" / "site.toml")

        # Initialize the different database objects
        self._static = DbsStatic(self)
        self._events = DbsEvent(
            self, standard_objects=self.site.standard_objects.events
        )
        self._scenarios = DbsScenario(self)
        self._strategies = DbsStrategy(
            self, standard_objects=self.site.standard_objects.strategies
        )
        self._measures = DbsMeasure(self)
        self._projections = DbsProjection(
            self, standard_objects=self.site.standard_objects.projections
        )
        self._benefits = DbsBenefit(self)

        # Delete any unfinished/crashed scenario output
        self.cleanup()

        self._init_done = True

    def shutdown(self):
        """Explicitly shut down the singleton and clear all references."""
        import gc

        self._instance = None
        self._init_done = False

        self.__class__._instance = None
        self.__dict__.clear()
        gc.collect()

    # Property methods
    @property
    def site(self) -> Site:
        return self._site

    @property
    def static(self) -> DbsStatic:
        return self._static

    @property
    def events(self) -> DbsEvent:
        return self._events

    @property
    def scenarios(self) -> DbsScenario:
        return self._scenarios

    @property
    def strategies(self) -> DbsStrategy:
        return self._strategies

    @property
    def measures(self) -> DbsMeasure:
        return self._measures

    @property
    def projections(self) -> DbsProjection:
        return self._projections

    @property
    def benefits(self) -> DbsBenefit:
        return self._benefits

    def interp_slr(self, slr_scenario: str, year: float) -> float:
        """Interpolate SLR value and reference it to the SLR reference year from the site toml.

        Parameters
        ----------
        slr_scenario : str
            SLR scenario name from the coulmn names in static/slr/slr.csv
        year : float
            year to evaluate

        Returns
        -------
        float
            _description_

        Raises
        ------
        ValueError
            if the reference year is outside of the time range in the slr.csv file
        ValueError
            if the year to evaluate is outside of the time range in the slr.csv file
        """
        if self.site.sfincs.slr_scenarios is None:
            raise ValueError("No SLR scenarios defined in the site configuration.")

        input_file = self.static_path / self.site.sfincs.slr_scenarios.file
        df = pd.read_csv(input_file)
        if year > df["year"].max() or year < df["year"].min():
            raise ValueError(
                "The selected year is outside the range of the available SLR scenarios"
            )
        else:
            slr = np.interp(year, df["year"], df[slr_scenario])
            ref_year = self.site.sfincs.slr_scenarios.relative_to_year
            if ref_year > df["year"].max() or ref_year < df["year"].min():
                raise ValueError(
                    f"The reference year {ref_year} is outside the range of the available SLR scenarios"
                )
            else:
                ref_slr = np.interp(ref_year, df["year"], df[slr_scenario])
                new_slr = us.UnitfulLength(
                    value=slr - ref_slr,
                    units=df["units"][0],
                )
                gui_units = self.site.gui.units.default_length_units
                return np.round(new_slr.convert(gui_units), decimals=2)

    # TODO: should probably be moved to frontend
    def plot_slr_scenarios(self) -> str:
        if self.site.sfincs.slr_scenarios is None:
            raise ValueError("No SLR scenarios defined in the site configuration.")
        input_file = self.input_path.parent.joinpath(
            "static", self.site.sfincs.slr_scenarios.file
        )
        df = pd.read_csv(input_file)
        ncolors = len(df.columns) - 2
        if "units" not in df.columns:
            raise ValueError(f"Expected column `units` in {input_file}.")

        units = df["units"].iloc[0]
        units = us.UnitTypesLength(units)

        if "Year" not in df.columns:
            if "year" not in df.columns:
                raise ValueError(f"Expected column `year` in {input_file}.")
            else:
                df = df.rename(columns={"year": "Year"})

        ref_year = self.site.sfincs.slr_scenarios.relative_to_year
        if ref_year > df["Year"].max() or ref_year < df["Year"].min():
            raise ValueError(
                f"The reference year {ref_year} is outside the range of the available SLR scenarios"
            )
        else:
            scenarios = self._static.get_slr_scn_names()
            for scn in scenarios:
                ref_slr = np.interp(ref_year, df["Year"], df[scn])
                df[scn] -= ref_slr

        df = df.drop(columns="units").melt(id_vars=["Year"]).reset_index(drop=True)
        # convert to units used in GUI
        slr_current_units = us.UnitfulLength(value=1.0, units=units)
        conversion_factor = slr_current_units.convert(
            self.site.gui.units.default_length_units
        )
        df.iloc[:, -1] = (conversion_factor * df.iloc[:, -1]).round(decimals=2)

        # rename column names that will be shown in html
        df = df.rename(
            columns={
                "variable": "Scenario",
                "value": f"Sea level rise [{self.site.gui.units.default_length_units.value}]",
            }
        )

        colors = sample_colorscale(
            "rainbow", [n / (ncolors - 1) for n in range(ncolors)]
        )
        fig = line(
            df,
            x="Year",
            y=f"Sea level rise [{self.site.gui.units.default_length_units.value}]",
            color="Scenario",
            color_discrete_sequence=colors,
        )

        # fig.update_traces(marker={"line": {"color": "#000000", "width": 2}})

        fig.update_layout(
            autosize=False,
            height=100 * 1.2,
            width=280 * 1.3,
            margin={"r": 0, "l": 0, "b": 0, "t": 0},
            font={"size": 10, "color": "black", "family": "Arial"},
            title_font={"size": 10, "color": "black", "family": "Arial"},
            legend_font={"size": 10, "color": "black", "family": "Arial"},
            legend_grouptitlefont={"size": 10, "color": "black", "family": "Arial"},
            legend={"entrywidthmode": "fraction", "entrywidth": 0.2},
            yaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
            xaxis_title=None,
            xaxis_range=[ref_year, df["Year"].max()],
            legend_title=None,
            # paper_bgcolor="#3A3A3A",
            # plot_bgcolor="#131313",
        )

        # write html to results folder
        output_loc = self.input_path.parent.joinpath("temp", "slr.html")
        output_loc.parent.mkdir(parents=True, exist_ok=True)
        fig.write_html(output_loc)
        return str(output_loc)

    def write_to_csv(self, name: str, event: Event, df: pd.DataFrame):
        df.to_csv(
            self.events.input_path.joinpath(event.name, f"{name}.csv"),
            header=False,
        )

    def write_cyc(self, event: Event, track: TropicalCyclone):
        cyc_file = self.events.input_path / event.name / f"{event.track_name}.cyc"
        # cht_cyclone function to write TropicalCyclone as .cyc file
        track.write_track(filename=cyc_file, fmt="ddb_cyc")

    def check_benefit_scenarios(self, benefit: Benefit) -> pd.DataFrame:
        """Return a dataframe with the scenarios needed for this benefit assessment run.

        Parameters
        ----------
        benefit : Benefit
        """
        runner = BenefitRunner(self, benefit=benefit)
        return runner.check_scenarios()

    def create_benefit_scenarios(self, benefit: Benefit) -> None:
        """Create any scenarios that are needed for the (cost-)benefit assessment and are not there already.

        Parameters
        ----------
        benefit : Benefit
        """
        runner = BenefitRunner(self, benefit=benefit)
        runner.check_scenarios()

        # Iterate through the scenarios needed and create them if not existing
        for index, row in runner.scenarios.iterrows():
            if row["scenario created"] == "No":
                scenario_dict = {}
                scenario_dict["event"] = row["event"]
                scenario_dict["projection"] = row["projection"]
                scenario_dict["strategy"] = row["strategy"]
                scenario_dict["name"] = "_".join(
                    [row["projection"], row["event"], row["strategy"]]
                )

                scenario_obj = Scenario(**scenario_dict)
                # Check if scenario already exists (because it was created before in the loop)
                try:
                    self.scenarios.save(scenario_obj)
                except ValueError as e:
                    if "name is already used" not in str(e):
                        # some other error was raised, so we re-raise it
                        raise e
                    # otherwise, if it already exists and we dont need to save it, we can just continue

        # Update the scenarios check
        runner.check_scenarios()

    def run_benefit(self, benefit_name: Union[str, list[str]]) -> None:
        """Run a (cost-)benefit analysis.

        Parameters
        ----------
        benefit_name : Union[str, list[str]]
            name(s) of the benefits to run.
        """
        if not isinstance(benefit_name, list):
            benefit_name = [benefit_name]
        for name in benefit_name:
            benefit = self.benefits.get(name)
            runner = BenefitRunner(self, benefit=benefit)
            runner.run_cost_benefit()

    def get_outputs(self) -> dict[str, Any]:
        """Return a dictionary with info on the outputs that currently exist in the database.

        Returns
        -------
        dict[str, Any]
            Includes 'name', 'path', 'last_modification_date' and "finished" info
        """
        all_scenarios = pd.DataFrame(self._scenarios.summarize_objects())
        if len(all_scenarios) > 0:
            df = all_scenarios[all_scenarios["finished"]]
        else:
            df = all_scenarios
        finished = df.drop(columns="finished").reset_index(drop=True)
        return finished.to_dict()

    def get_topobathy_path(self) -> str:
        """Return the path of the topobathy tiles in order to create flood maps with water level maps.

        Returns
        -------
        str
            path to topobathy tiles
        """
        path = self.input_path.parent.joinpath("static", "dem", "tiles", "topobathy")
        return str(path)

    def get_index_path(self) -> str:
        """Return the path of the index tiles which are used to connect each water level cell with the topobathy tiles.

        Returns
        -------
        str
            path to index tiles
        """
        path = self.input_path.parent.joinpath("static", "dem", "tiles", "indices")
        return str(path)

    def get_depth_conversion(self) -> float:
        """Return the flood depth conversion that is need in the gui to plot the flood map.

        Returns
        -------
        float
            conversion factor
        """
        # Get conresion factor need to get from the sfincs units to the gui units
        units = us.UnitfulLength(
            value=1, units=self.site.gui.units.default_length_units
        )
        unit_cor = units.convert(new_units=us.UnitTypesLength.meters)

        return unit_cor

    def get_max_water_level(
        self,
        scenario_name: str,
        return_period: Optional[int] = None,
    ) -> np.ndarray:
        """Return an array with the maximum water levels during an event.

        Parameters
        ----------
        scenario_name : str
            name of scenario
        return_period : int, optional
            return period in years, by default None

        Returns
        -------
        np.array
            2D map of maximum water levels
        """
        # If single event read with hydromt-sfincs
        if not return_period:
            map_path = self.scenarios.output_path.joinpath(
                scenario_name,
                "Flooding",
                "max_water_level_map.nc",
            )
            with xr.open_dataarray(map_path) as map:
                zsmax = map.to_numpy()
        else:
            file_path = self.scenarios.output_path.joinpath(
                scenario_name,
                "Flooding",
                f"RP_{return_period:04d}_maps.nc",
            )
            with xr.open_dataset(file_path) as ds:
                zsmax = ds["risk_map"][:, :].to_numpy().T
        return zsmax

    def get_building_footprints(self, scenario_name: str) -> GeoDataFrame:
        """Return a geodataframe of the impacts at the footprint level.

        Parameters
        ----------
        scenario_name : str
            name of scenario

        Returns
        -------
        GeoDataFrame
            impacts at footprint level
        """
        out_path = self.scenarios.output_path.joinpath(scenario_name, "Impacts")
        footprints = out_path / f"Impacts_building_footprints_{scenario_name}.gpkg"
        gdf = gpd.read_file(footprints, engine="pyogrio")
        gdf = gdf.to_crs(4326)
        return gdf

    def get_roads(self, scenario_name: str) -> GeoDataFrame:
        """Return a geodataframe of the impacts at roads.

        Parameters
        ----------
        scenario_name : str
            name of scenario

        Returns
        -------
        GeoDataFrame
            Impacts at roads
        """
        out_path = self.scenarios.output_path.joinpath(scenario_name, "Impacts")
        roads = out_path / f"Impacts_roads_{scenario_name}.gpkg"
        gdf = gpd.read_file(roads, engine="pyogrio")
        gdf = gdf.to_crs(4326)
        return gdf

    def get_aggregation(self, scenario_name: str) -> dict[str, gpd.GeoDataFrame]:
        """Return a dictionary with the aggregated impacts as geodataframes.

        Parameters
        ----------
        scenario_name : str
            name of the scenario

        Returns
        -------
        dict[GeoDataFrame]
            dictionary with aggregated damages per aggregation type
        """
        out_path = self.scenarios.output_path.joinpath(scenario_name, "Impacts")
        gdfs = {}
        for aggr_area in out_path.glob(f"Impacts_aggregated_{scenario_name}_*.gpkg"):
            label = aggr_area.stem.split(f"{scenario_name}_")[-1]
            gdfs[label] = gpd.read_file(aggr_area, engine="pyogrio")
            gdfs[label] = gdfs[label].to_crs(4326)
        return gdfs

    def get_aggregation_benefits(
        self, benefit_name: str
    ) -> dict[str, gpd.GeoDataFrame]:
        """Get a dictionary with the aggregated benefits as geodataframes.

        Parameters
        ----------
        benefit_name : str
            name of the benefit analysis

        Returns
        -------
        dict[GeoDataFrame]
            dictionary with aggregated benefits per aggregation type
        """
        out_path = self.benefits.output_path.joinpath(
            benefit_name,
        )
        gdfs = {}
        for aggr_area in out_path.glob("benefits_*.gpkg"):
            label = aggr_area.stem.split("benefits_")[-1]
            gdfs[label] = gpd.read_file(aggr_area, engine="pyogrio")
            gdfs[label] = gdfs[label].to_crs(4326)
        return gdfs

    def get_object_list(
        self,
        object_type: Literal[
            "projections", "events", "measures", "strategies", "scenarios", "benefits"
        ],
    ) -> dict[str, Any]:
        """Get a dictionary with all the toml paths and last modification dates that exist in the database that correspond to object_type.

        Parameters
        ----------
        object_type : str
            Can be 'projections', 'events', 'measures', 'strategies' or 'scenarios'

        Returns
        -------
        dict[str, Any]
            Includes 'path' and 'last_modification_date' info
        """
        match object_type:
            case "projections":
                return self.projections.summarize_objects()
            case "events":
                return self.events.summarize_objects()
            case "measures":
                return self.measures.summarize_objects()
            case "strategies":
                return self.strategies.summarize_objects()
            case "scenarios":
                return self.scenarios.summarize_objects()
            case "benefits":
                return self.benefits.summarize_objects()
            case _:
                raise ValueError(
                    f"Object type '{object_type}' is not valid. Must be one of 'projections', 'events', 'measures', 'strategies' or 'scenarios'."
                )

    def has_run_hazard(self, scenario_name: str) -> None:
        """Check if there is already a simulation that has the exact same hazard component.

        If yes that is copied to avoid running the hazard model twice.

        Parameters
        ----------
        scenario_name : str
            name of the scenario to check if needs to be rerun for hazard
        """
        scenario = self.scenarios.get(scenario_name)
        runner = ScenarioRunner(self, scenario=scenario)

        # Dont do anything if the hazard model has already been run in itself
        if runner.impacts.hazard.has_run:
            return

        scenarios = [
            self.scenarios.get(scn)
            for scn in self.scenarios.summarize_objects()["name"]
        ]
        scns_simulated = [
            sim
            for sim in scenarios
            if self.scenarios.output_path.joinpath(sim.name, "Flooding").is_dir()
        ]

        for scn in scns_simulated:
            if self.scenarios.equal_hazard_components(scn, scenario):
                existing = self.scenarios.output_path.joinpath(scn.name, "Flooding")
                path_new = self.scenarios.output_path.joinpath(
                    scenario.name, "Flooding"
                )
                _runner = ScenarioRunner(self, scenario=scn)

                if _runner.impacts.hazard.has_run:  # only copy results if the hazard model has actually finished and skip simulation folders
                    shutil.copytree(
                        existing,
                        path_new,
                        dirs_exist_ok=True,
                        ignore=shutil.ignore_patterns("simulations"),
                    )
                    self.logger.info(
                        f"Hazard simulation is used from the '{scn.name}' scenario"
                    )

    def cleanup(self) -> None:
        """
        Remove corrupted scenario output.

        This method removes any scenario output that:
            - is corrupted due to unfinished runs
            - does not have a corresponding input

        """
        if not Settings().delete_crashed_runs:
            return

        if not self.scenarios.output_path.is_dir():
            return

        input_scenarios = [
            (self.scenarios.input_path / dir).resolve()
            for dir in os.listdir(self.scenarios.input_path)
        ]
        output_scenarios = [
            (self.scenarios.output_path / dir).resolve()
            for dir in os.listdir(self.scenarios.output_path)
        ]

        def _call_garbage_collector(func, path, exc_info, retries=5, delay=0.1):
            """Retry deletion up to 5 times if the file is locked."""
            for attempt in range(retries):
                gc.collect()
                time.sleep(delay)
                try:
                    func(path)  # Retry deletion
                    return  # Exit if successful
                except Exception as e:
                    print(
                        f"Attempt {attempt + 1}/{retries} failed to delete {path}: {e}"
                    )

            print(f"Giving up on deleting {path} after {retries} attempts.")

        for dir in output_scenarios:
            # Delete if: input was deleted or corrupted output due to unfinished run
            if dir.name not in [
                path.name for path in input_scenarios
            ] or not finished_file_exists(dir):
                shutil.rmtree(dir, onerror=_call_garbage_collector)



import platform

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest

from flood_adapt.flood_adapt import FloodAdapt
from flood_adapt.misc.utils import finished_file_exists
from flood_adapt.objects.events.hurricane import HurricaneEvent
from flood_adapt.objects.measures.measures import Measure
from flood_adapt.objects.scenarios.scenarios import Scenario
from flood_adapt.workflows.benefit_runner import BenefitRunner
from tests.data.create_test_input import create_event_set_with_hurricanes
from tests.test_adapter.test_sfincs_adapter import mock_meteohandler_read
from tests.test_objects.test_events.test_eventset import (
    test_eventset,
    test_sub_event,
)
from tests.test_objects.test_events.test_historical import (
    setup_nearshore_event,
    setup_offshore_meteo_event,
    setup_offshore_scenario,
)
from tests.test_objects.test_events.test_hurricane import setup_hurricane_event
from tests.test_objects.test_events.test_synthetic import test_event_all_synthetic
from tests.test_objects.test_measures.test_measures import (
    test_buyout,
    test_elevate,
    test_floodproof,
    test_floodwall,
    test_green_infra,
    test_pump,
)

# To stop ruff from deleting these 'unused' imports
__all__ = [
    # Events
    "test_eventset",
    "test_sub_event",
    "test_event_all_synthetic",
    "create_event_set_with_hurricanes",
    "setup_nearshore_event",
    "setup_offshore_meteo_event",
    "setup_hurricane_event",
    # Scenarios
    "setup_offshore_scenario",
    # Mock
    "mock_meteohandler_read",
    # Measures
    "test_buyout",
    "test_elevate",
    "test_floodproof",
    "test_floodwall",
    "test_pump",
    "test_green_infra",
]


@pytest.fixture(scope="session")
def get_rng():
    yield np.random.default_rng(2021)


class TestEvents:
    @pytest.fixture()
    def test_dict(self):
        test_dict = {
            "name": "extreme12ft",
            "description": "extreme 12 foot event",
            "mode": "single_event",
            "template": "Synthetic",
            "timing": "idealized",
            "water_level_offset": {"value": 0, "units": "feet"},
            "wind": {
                "source": "constant",
                "constant_speed": {"value": 0, "units": "m/s"},
                "constant_direction": {"value": 0, "units": "deg N"},
            },
            "rainfall": {"source": "none"},
            "river": [
                {
                    "source": "constant",
                    "constant_discharge": {"value": 5000, "units": "cfs"},
                }
            ],
            "time": {"duration_before_t0": 24, "duration_after_t0": "24"},
            "tide": {
                "source": "harmonic",
                "harmonic_amplitude": {"value": 3, "units": "feet"},
            },
            "surge": {
                "source": "shape",
                "shape_type": "gaussian",
                "shape_duration": 24,
                "shape_peak_time": 0,
                "shape_peak": {"value": 9.22, "units": "feet"},
            },
        }
        yield test_dict

    def test_create_synthetic_event_valid_dict(self, test_fa: FloodAdapt, test_dict):
        # When user presses add event and chooses the events
        # the dictionary is returned and an Event object is created
        test_fa.create_event(test_dict)
        # TODO assert event attrs

    def test_create_synthetic_event_invalid_dict(self, test_fa: FloodAdapt, test_dict):
        del test_dict["name"]
        with pytest.raises(ValueError):
            # Assert error if a value is incorrect
            test_fa.create_event(test_dict)
        # TODO assert error msg

    def test_save_synthetic_event_already_exists(self, test_fa: FloodAdapt, test_dict):
        event = test_fa.create_event(test_dict)
        if test_dict["name"] not in test_fa.get_events()["name"]:
            test_fa.save_event(event)

        with pytest.raises(ValueError):
            test_fa.save_event(event)
        # TODO assert error msg

    def test_save_event_valid(self, test_fa: FloodAdapt, test_dict):
        # Change name to something new
        test_dict["name"] = "testNew"
        event = test_fa.create_event(test_dict)
        if test_dict["name"] in test_fa.get_events()["name"]:
            test_fa.delete_event(test_dict["name"])
        test_fa.save_event(event)
        # TODO assert event attrs

    def test_delete_event_doesnt_exist(self, test_fa: FloodAdapt):
        # apparently this doesnt raise an error?
        test_fa.delete_event("doesnt_exist")


class TestProjections:
    def test_projection(self, test_fa: FloodAdapt):
        test_dict = {
            "name": "SLR_2ft",
            "description": "SLR_2ft",
            "physical_projection": {
                "sea_level_rise": {"value": "two", "units": "feet"},
                "subsidence": {"value": 1, "units": "feet"},
            },
            "socio_economic_change": {},
        }
        # When user presses add projection and chooses the projections
        # the dictionary is returned and an Projection object is created
        with pytest.raises(ValueError):
            # Assert error if a value is incorrect
            projection = test_fa.create_projection(test_dict)

        # correct projection
        test_dict["physical_projection"]["sea_level_rise"]["value"] = 2
        projection = test_fa.create_projection(test_dict)

        with pytest.raises(ValueError):
            # Assert error if name already exists
            test_fa.save_projection(projection)

        # Change name to something new
        test_dict["name"] = "test_proj_1"
        projection = test_fa.create_projection(test_dict)
        # If the name is not used before the measure is save in the database
        test_fa.save_projection(projection)
        test_fa.database.projections.summarize_objects()

        # Try to delete a measure which is already used in a scenario
        # with pytest.raises(ValueError):
        #    api_projections.delete_measure("", database)

        # If user presses delete projection the measure is deleted
        test_fa.delete_projection("test_proj_1")
        test_fa.database.projections.summarize_objects()


class TestMeasures:
    # dict of measure fixture names and their corresponding measure type
    measure_fixtures = {
        "test_elevate": "elevate_properties",
        "test_buyout": "buyout_properties",
        "test_floodproof": "floodproof_properties",
        "test_floodwall": "floodwall",
        "test_pump": "pump",
        "test_green_infra": "greening",
    }

    @pytest.mark.parametrize("measure_fixture_name", measure_fixtures.keys())
    def test_create_measure(
        self, test_fa_class: FloodAdapt, measure_fixture_name, request
    ):
        measure: Measure = request.getfixturevalue(measure_fixture_name)
        measure = test_fa_class.create_measure(
            attrs=measure.model_dump(exclude_none=True), type=measure.type
        )
        assert measure is not None

    @pytest.mark.parametrize("measure_fixture", measure_fixtures.keys())
    def test_save_measure(self, test_fa: FloodAdapt, measure_fixture, request):
        measure = request.getfixturevalue(measure_fixture)

        test_fa.save_measure(measure)
        assert (test_fa.database.measures.input_path / measure.name).exists()

    @pytest.mark.parametrize("measure_fixture", measure_fixtures.keys())
    def test_get_measure(self, test_fa: FloodAdapt, measure_fixture, request):
        measure = request.getfixturevalue(measure_fixture)

        test_fa.save_measure(measure)
        assert (test_fa.database.measures.input_path / measure.name).exists()

        loaded_measure = test_fa.get_measure(measure.name)
        assert loaded_measure == measure

    @pytest.mark.parametrize("measure_fixture", measure_fixtures.keys())
    def test_delete_measure(self, test_fa: FloodAdapt, measure_fixture, request):
        measure = request.getfixturevalue(measure_fixture)
        test_fa.save_measure(measure)
        assert (test_fa.database.measures.input_path / measure.name).exists()

        test_fa.delete_measure(measure.name)
        assert not (test_fa.database.measures.input_path / measure.name).exists()

    @pytest.mark.parametrize("measure_fixture", measure_fixtures.keys())
    def test_copy_measure(self, test_fa: FloodAdapt, measure_fixture, request):
        measure = request.getfixturevalue(measure_fixture)
        test_fa.save_measure(measure)
        assert (test_fa.database.measures.input_path / measure.name).exists()

        new_name = f"copy_{measure.name}"
        new_description = f"copy of {measure.description}"

        test_fa.copy_measure(
            old_name=measure.name, new_name=new_name, new_description=new_description
        )
        new_measure = test_fa.get_measure(new_name)

        assert (test_fa.database.measures.input_path / new_name).exists()
        assert measure == new_measure


class TestStrategies:
    def test_strategy(self, test_fa: FloodAdapt):
        strat_with_existing_name = {
            "name": "strategy_comb",
            "description": "strategy_comb",
            "measures": [
                "seawall",
                "raise_property_aggregation_area",
                "raise_property_polygon",
            ],
        }
        # Create a new strategy object with a name that already exists
        strategy = test_fa.create_strategy(strat_with_existing_name)

        # Save it in the database -> name exists error
        with pytest.raises(ValueError):
            test_fa.save_strategy(strategy)

        # Delete a strategy which is already used in a scenario
        with pytest.raises(ValueError):
            test_fa.delete_strategy("strategy_comb")

        # Change to unused name
        strategy.name = "test_strat_1"

        test_fa.save_strategy(strategy)
        assert test_fa.get_strategy(strategy.name) == strategy

        test_fa.delete_strategy(strategy.name)
        with pytest.raises(ValueError):
            test_fa.get_strategy(strategy.name)


@pytest.mark.skipif(
    platform.system() == "Linux",
    reason="Skipped on Linux due to broken sfincs binary",
)
class TestScenarios:
    @pytest.fixture()
    def setup_nearshore_scenario(self, test_fa: FloodAdapt, setup_nearshore_event):
        test_fa.save_event(setup_nearshore_event)

        scn = Scenario(
            name="gauged_nearshore",
            description="current_extreme12ft_no_measures",
            event=setup_nearshore_event.name,
            projection="current",
            strategy="no_measures",
        )
        return scn

    @pytest.fixture()
    def setup_offshore_meteo_scenario(
        self,
        test_fa: FloodAdapt,
        setup_offshore_meteo_event,
        mock_meteohandler_read,
    ):
        test_fa.save_event(setup_offshore_meteo_event)

        scn = Scenario(
            name="offshore_meteo",
            event=setup_offshore_meteo_event.name,
            projection="current",
            strategy="no_measures",
        )

        return scn

    @pytest.fixture()
    def setup_hurricane_scenario(
        self,
        test_fa: FloodAdapt,
        setup_hurricane_event: HurricaneEvent,
        mock_meteohandler_read,
    ) -> tuple[Scenario, HurricaneEvent]:
        event = setup_hurricane_event
        scn = Scenario(
            name="hurricane",
            event=event.name,
            projection="current",
            strategy="no_measures",
        )
        test_fa.save_event(event)
        test_fa.save_scenario(scn)
        return scn, event

    @pytest.fixture()
    def setup_synthetic_scenario(self, test_fa: FloodAdapt, test_event_all_synthetic):
        test_fa.save_event(test_event_all_synthetic)

        scn = Scenario(
            name="synthetic",
            event=test_event_all_synthetic.name,
            projection="current",
            strategy="no_measures",
        )
        return scn

    @pytest.fixture()
    def setup_eventset_scenario(
        self, test_fa: FloodAdapt, dummy_projection, dummy_strategy, test_eventset
    ):
        test_fa.save_projection(dummy_projection)
        test_fa.save_strategy(dummy_strategy)
        test_fa.save_event(test_eventset)

        scn = Scenario(
            name="test_risk_scenario_with_hurricanes",
            event=test_eventset.name,
            projection=dummy_projection.name,
            strategy=dummy_strategy.name,
        )
        return test_fa, scn, test_eventset

    def test_run_offshore_scenario(
        self, test_fa: FloodAdapt, setup_offshore_meteo_scenario
    ):
        test_fa.save_scenario(setup_offshore_meteo_scenario)
        test_fa.run_scenario(setup_offshore_meteo_scenario.name)

        assert finished_file_exists(
            test_fa.database.scenarios.output_path / setup_offshore_meteo_scenario.name
        )

    def test_run_nearshore_scenario(
        self, test_fa: FloodAdapt, setup_nearshore_scenario
    ):
        test_fa.save_scenario(setup_nearshore_scenario)
        test_fa.run_scenario(setup_nearshore_scenario.name)

        assert finished_file_exists(
            test_fa.database.scenarios.output_path / setup_nearshore_scenario.name
        )

    def test_run_synthetic_scenario(
        self, test_fa: FloodAdapt, setup_synthetic_scenario
    ):
        test_fa.save_scenario(setup_synthetic_scenario)
        test_fa.run_scenario(setup_synthetic_scenario.name)

        assert finished_file_exists(
            test_fa.database.scenarios.output_path / setup_synthetic_scenario.name
        )

    def test_run_hurricane_scenario(
        self, test_fa: FloodAdapt, setup_hurricane_scenario
    ):
        scn, event = setup_hurricane_scenario
        test_fa.save_scenario(scn)
        test_fa.run_scenario(scn.name)

        assert finished_file_exists(test_fa.database.scenarios.output_path / scn.name)

    def test_run_eventset_scenario(self, test_fa: FloodAdapt, setup_eventset_scenario):
        test_fa, scn, event_set = setup_eventset_scenario
        test_fa.save_scenario(scn)
        test_fa.run_scenario(scn.name)

        assert finished_file_exists(test_fa.database.scenarios.output_path / scn.name)

    def test_create_save_scenario(
        self, test_fa: FloodAdapt, setup_offshore_meteo_event
    ):
        test_fa.save_event(setup_offshore_meteo_event)

        test_dict = {
            "name": "current_extreme12ft_no_measures",
            "description": "current_extreme12ft_no_measures",
            "projection": "current",
            "strategy": "no_measures",
        }
        # When user presses add scenario and chooses the measures
        # the dictionary is returned and a Strategy object is created
        with pytest.raises(ValueError):
            # Assert error if the event is not a correct value
            scenario = test_fa.create_scenario(test_dict)

        # correct event
        test_dict["event"] = setup_offshore_meteo_event.name
        scenario = test_fa.create_scenario(test_dict)

        assert not test_fa.save_scenario(scenario)[0]

        # Change name to something new
        test_dict["name"] = "test1"
        scenario = test_fa.create_scenario(test_dict)
        test_fa.save_scenario(scenario)
        test_fa.database.scenarios.summarize_objects()

        # If user presses delete scenario the measure is deleted
        test_fa.delete_scenario("test1")
        test_fa.database.scenarios.summarize_objects()


@pytest.mark.skipif(
    platform.system() == "Linux",
    reason="Skipped on Linux due to broken sfincs binary",
)
class TestOutput:
    @pytest.fixture(scope="class")
    def completed_scenario(self, test_fa_class: FloodAdapt):
        scenario_name = "current_extreme12ft_no_measures"
        test_fa_class.run_scenario(scenario_name)
        yield test_fa_class, scenario_name

    def test_impact_metrics(self, completed_scenario: tuple[FloodAdapt, str]):
        test_fa, scn_name = completed_scenario
        metrics = test_fa.get_infometrics(scn_name)
        assert isinstance(metrics, pd.DataFrame)

    def test_impact_footprints(self, completed_scenario: tuple[FloodAdapt, str]):
        test_fa, scn_name = completed_scenario
        footprints = test_fa.get_building_footprint_impacts(scn_name)
        assert isinstance(footprints, gpd.GeoDataFrame)

    def test_impact_get_aggregation(self, completed_scenario: tuple[FloodAdapt, str]):
        test_fa, scn_name = completed_scenario
        aggr_areas = test_fa.get_aggregated_impacts(scn_name)
        assert isinstance(aggr_areas, dict)


class TestStatic:
    def test_buildings(self, test_fa: FloodAdapt):
        assert isinstance(test_fa.get_building_geometries(), gpd.GeoDataFrame)

    def test_aggr_areas(self, test_fa: FloodAdapt):
        aggr_areas = test_fa.get_aggregation_areas()

        assert isinstance(aggr_areas, dict)
        assert isinstance(aggr_areas["aggr_lvl_1"], gpd.GeoDataFrame)

    def test_building_types(self, test_fa: FloodAdapt):
        types = test_fa.get_building_types()
        expected_types = ["RES", "COM", "all"]

        assert isinstance(types, list)
        assert len(types) == 3
        assert all(t in types for t in expected_types)


class TestBenefits:
    @pytest.mark.skip(reason="fix this test")
    def test_benefit(self, test_fa: FloodAdapt, get_rng):
        # Inputs for benefit calculation
        # Name given already exists to do test for error capture
        config = test_fa.database.site.fiat.benefits
        benefit_dict = {
            "name": "benefit_raise_properties_2050",
            "description": "",
            "event_set": "test_set",
            "strategy": "elevate_comb_correct",
            "projection": "all_projections",
            "future_year": "two thousand eighty",
            "current_situation": {
                "projection": config.current_projection,
                "year": config.current_year,
            },
            "baseline_strategy": config.baseline_strategy,
            "discount_rate": 0.07,
        }

        # When user tries to create benefits calculation, it will return an error since year has wrong format
        with pytest.raises(ValueError):
            benefit = test_fa.create_benefit(benefit_dict)

        # correct value of year
        benefit_dict["future_year"] = 2080
        benefit = test_fa.create_benefit(benefit_dict)

        # When user tries to create benefits calculation, it will return an error since name already exists
        with pytest.raises(ValueError):
            test_fa.save_benefit(benefit)

        # Change name to something new
        benefit_dict["name"] = "benefit_raise_properties_2080"
        benefit = test_fa.create_benefit(benefit_dict)

        # When user presses "Check scenarios" the object will be created in memory and a dataframe will be returned
        df = test_fa.check_benefit_scenarios(benefit)

        # If not all scenarios are there you should get an error
        if sum(df["scenario created"] == "No") > 0:
            with pytest.raises(ValueError):
                test_fa.save_benefit(benefit)

        # Create missing scenarios
        test_fa.create_benefit_scenarios(benefit)

        # Save benefit calculation
        df = test_fa.check_benefit_scenarios(benefit)
        if sum(df["scenario created"] == "No") == 0:
            test_fa.save_benefit(benefit)

        # Get error when the scenarios are not run
        df = test_fa.check_benefit_scenarios(benefit)
        if not all(df["scenario run"]):
            with pytest.raises(RuntimeError):
                # Assert error if not yet run
                test_fa.run_benefit("benefit_raise_properties_2080")

        # Create dummy data
        aggrs = test_fa.get_aggregation_areas()
        damages_dummy = {
            "current_no_measures": 100e6,
            "current_with_strategy": 50e6,
            "future_no_measures": 300e6,
            "future_with_strategy": 180e6,
        }

        runner = BenefitRunner(test_fa.database, benefit)

        for name, row in runner.scenarios.iterrows():
            # Create output folder
            output_path = (
                test_fa.database.output_path / "scenarios" / row["scenario created"]
            )
            if not output_path.exists():
                output_path.mkdir(parents=True)
            # Create dummy metrics file
            dummy_metrics = pd.DataFrame(
                {
                    "Description": "",
                    "Show In Metrics Table": "TRUE",
                    "Long Name": "",
                    "Value": damages_dummy[name],
                },
                index=["ExpectedAnnualDamages"],
            )

            dummy_metrics.to_csv(
                output_path.joinpath(f"Infometrics_{row['scenario created']}.csv")
            )

            # Create dummy metrics for aggregation areas
            for aggr_type, aggr in aggrs.items():
                # Generate random distribution of damage per aggregation area
                rng = np.random.default_rng(seed=2024)
                dmgs = rng.random(len(aggr))
                dmgs = dmgs / dmgs.sum() * damages_dummy[name]

                dict0 = {
                    "Description": ["", ""],
                    "Show In Metrics Table": ["TRUE", "TRUE"],
                    "Long Name": ["", ""],
                }

                for i, aggr_area in enumerate(aggr["name"]):
                    dict0[aggr_area] = [dmgs[i], get_rng.normal(1, 0.2) * dmgs[i]]

                dummy_metrics_aggr = pd.DataFrame(
                    dict0, index=["ExpectedAnnualDamages", "EWEAD"]
                ).T

                dummy_metrics_aggr.to_csv(
                    output_path.joinpath(
                        f"Infometrics_{row['scenario created']}_{aggr_type}.csv"
                    )
                )
            fiat_path = output_path.joinpath("Impacts", "fiat_model", "fiat.log")
            fiat_path.parent.mkdir(parents=True, exist_ok=True)
            f = open(fiat_path, "w")
            f.write("Geom calculation are done!")
            f.close()

        test_fa.run_benefit("benefit_raise_properties_2080")

        benefit = test_fa.get_benefit("benefit_raise_properties_2080")
        runner = BenefitRunner(test_fa.database, benefit)
        assert (
            len(runner.results.keys()) == 2
        )  # check if only benefits and html are produced

        # If the user edits the benefit to add costs
        benefit_dict_new = benefit_dict.copy()
        benefit_dict_new["implementation_cost"] = 30000000
        benefit_dict_new["annual_maint_cost"] = 0
        benefit = test_fa.create_benefit(benefit_dict_new)
        test_fa.save_benefit(benefit, overwrite=True)
        test_fa.run_benefit("benefit_raise_properties_2080")

        benefit = test_fa.get_benefit("benefit_raise_properties_2080")
        runner = BenefitRunner(test_fa.database, benefit)
        assert "costs" in runner.results.keys()  # check if costs is in results
        assert "BCR" in runner.results.keys()  # check if BCR is in results

        test_fa.delete_benefit("benefit_raise_properties_2080")



from flood_adapt.objects.object_model import Object


class Scenario(Object):
    """BaseModel describing the expected variables and data types of a scenario.

    A scenario is a combination of an event, a projection, and a strategy, that all should be saved in the database.

    Attributes
    ----------
    event : str
        The name of the event.
    projection : str
        The name of the projection.
    strategy : str
        The name of the strategy.

    """

    event: str
    projection: str
    strategy: str



from enum import Enum
from pathlib import Path
from typing import Optional

from flood_adapt.config.config import Settings


class TopLevelDir(str, Enum):
    """Top level directories in the database."""

    input = "input"
    output = "output"
    static = "static"
    temp = "temp"


class ObjectDir(str, Enum):
    """The names for object directories at the second level of the database."""

    site = "site"

    benefit = "benefits"
    event = "events"
    strategy = "strategies"
    measure = "measures"
    projection = "projections"
    scenario = "scenarios"
    config = "config"

    # buyout = "measures"
    # elevate = "measures"
    # floodproof = "measures"
    # greening = "measures"
    # floodwall = "measures"
    # pump = "measures"


def db_path(
    top_level_dir: TopLevelDir = TopLevelDir.input,
    object_dir: Optional[ObjectDir] = None,
    obj_name: Optional[str] = None,
) -> Path:
    """Return an path to a database directory from arguments."""
    rel_path = Path(top_level_dir.value)
    if object_dir is not None:
        if isinstance(object_dir, ObjectDir):
            rel_path = rel_path / object_dir.value
        else:
            rel_path = rel_path / str(object_dir)

        if obj_name is not None:
            rel_path = rel_path / obj_name

    return Settings().database_path / rel_path



import enum
import math
from abc import ABC
from datetime import timedelta
from enum import Enum
from typing import Generic, Optional, Type, TypeVar, Union

from pydantic import BaseModel, Field, model_validator

__all__ = [
    "ValueUnitPair",
    "VerticalReference",
    "UnitTypesLength",
    "UnitTypesArea",
    "UnitTypesVolume",
    "UnitTypesVelocity",
    "UnitTypesDirection",
    "UnitTypesTime",
    "UnitTypesDischarge",
    "UnitTypesIntensity",
    "UnitfulLength",
    "UnitfulHeight",
    "UnitfulLengthRefValue",
    "UnitfulArea",
    "UnitfulVelocity",
    "UnitfulDirection",
    "UnitfulDischarge",
    "UnitfulIntensity",
    "UnitfulVolume",
    "UnitfulTime",
    "ValueUnitPairs",
]

TUnit = TypeVar("TUnit", bound=enum.Enum)
TClass = TypeVar("TClass", bound="ValueUnitPair")


class ValueUnitPair(ABC, BaseModel, Generic[TUnit]):
    """
    Represents a value with associated units.

    Attributes
    ----------
    value: float
        The numerical value.
    units : Unit
        The units of the value.
    """

    _DEFAULT_UNIT: TUnit
    _CONVERSION_FACTORS: dict[TUnit, float]

    value: float
    units: TUnit

    def convert(self, new_units: TUnit) -> float:
        """Return the value converted to the new units.

        Parameters
        ----------
        new_units : Unit
            The new units.

        Returns
        -------
        converted_value : float
            The converted value.
        """
        if new_units not in self._CONVERSION_FACTORS:
            raise ValueError(f"Invalid units: {new_units}")
        in_default_units = self.value / self._CONVERSION_FACTORS[self.units]
        return in_default_units * self._CONVERSION_FACTORS[new_units]

    def transform(self: TClass, new_units: TUnit) -> TClass:
        """Return a new ValueUnitPair instance with the value converted to the new units.

        Parameters
        ----------
        new_units : Unit
            The new units.

        Returns
        -------
        value_unit_pair : ValueUnitPair
            The new ValueUnitPair instance with the value converted to the new units and the new units.
        """
        return type(self)(value=self.convert(new_units), units=new_units)

    @model_validator(mode="before")
    @classmethod
    def extract_unit_class(cls, data: Optional[dict]) -> "ValueUnitPair":
        if cls is not ValueUnitPair:
            return data  # Already in the right subclass, don't interfere

        if not isinstance(data, dict):
            raise TypeError("Expected dictionary for deserialization.")

        str_unit = data.get("units")
        if not str_unit:
            raise ValueError("Missing 'units' field in input data.")

        for unit_enum_cls, vu_cls in UNIT_TO_CLASS.items():
            try:
                enum_unit = unit_enum_cls(str_unit)
                data["units"] = enum_unit
                return vu_cls(**data)
            except ValueError:
                continue

        raise ValueError(f"Unsupported or unknown unit: {str_unit}")

    def __str__(self) -> str:
        return f"{self.value} {self.units.value}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}(value={self.value}, units={self.units})"

    def __sub__(self: TClass, other: TClass) -> TClass:
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Cannot compare self: {type(self).__name__} to other: {type(other).__name__}"
            )
        return type(self)(
            value=self.value - other.convert(self.units), units=self.units
        )

    def __add__(self: TClass, other: TClass) -> TClass:
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Cannot compare self: {type(self).__name__} to other: {type(other).__name__}"
            )
        return type(self)(
            value=self.value + other.convert(self.units), units=self.units
        )

    def __eq__(self: TClass, other: TClass) -> bool:
        if not isinstance(other, ValueUnitPair):
            raise TypeError(
                f"Cannot compare self: {type(self).__name__} to other: {type(other).__name__}"
            )
        if not (hasattr(other, "value") and hasattr(other, "units")):
            raise AttributeError(f"Incomplete UnitfulValue instance: {other}")

        if not isinstance(other.units, type(self.units)):
            raise TypeError(
                f"Cannot compare self: {type(self).__name__} to other: {type(other).__name__}"
            )

        return math.isclose(
            self.value, other.convert(self.units), rel_tol=1e-2
        )  # 1% relative tolerance for equality. So 1.0 == 1.01 evaluates to True

    def __lt__(self: TClass, other: TClass) -> bool:
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Cannot compare self: {type(self).__name__} to other: {type(other).__name__}"
            )
        return self.value < other.convert(self.units)

    def __gt__(self: TClass, other: TClass) -> bool:
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Cannot compare self: {type(self).__name__} to other: {type(other).__name__}"
            )
        return self.value > other.convert(self.units)

    def __ge__(self: TClass, other: TClass) -> bool:
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Cannot compare self: {type(self).__name__} to other: {type(other).__name__}"
            )
        return (self > other) or (self == other)

    def __le__(self: TClass, other: TClass) -> bool:
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Cannot compare self: {type(self).__name__} to other: {type(other).__name__}"
            )
        return (self < other) or (self == other)

    def __ne__(self: TClass, other: TClass) -> bool:
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Cannot compare self: {type(self).__name__} to other: {type(other).__name__}"
            )
        return not (self == other)

    def __mul__(self: TClass, other: int | float) -> TClass:
        if isinstance(other, int) or isinstance(other, float):
            return type(self)(value=self.value * other, units=self.units)
        else:
            raise TypeError(
                f"Cannot multiply self: {type(self).__name__} with other: {type(other).__name__}. Only int and float are allowed."
            )

    def __div__(self: TClass, other: TClass | int | float) -> TClass | float:
        if isinstance(other, int) or isinstance(other, float):
            return type(self)(value=self.value / other, units=self.units)
        elif isinstance(other, type(self)):
            return self.value / other.convert(self.units)
        else:
            raise TypeError(
                f"Cannot divide self: {type(self).__name__} with other: {type(other).__name__}. Only {type(self).__name__}, int and float are allowed."
            )

    def __truediv__(self: TClass, other: TClass | int | float) -> TClass | float:
        if isinstance(other, int) or isinstance(other, float):
            return type(self)(value=self.value / other, units=self.units)
        elif isinstance(other, type(self)):
            return self.value / other.convert(self.units)
        else:
            raise TypeError(
                f"Cannot divide self: {type(self).__name__} with other: {type(other).__name__}. Only {type(self).__name__}, int and float are allowed."
            )


class UnitTypesLength(str, Enum):
    """Units of length.

    Attributes
    ----------
    meters : meters
    centimeters : centimeters
    millimeters : millimeters
    feet : feet
    inch : inch
    miles : miles
    """

    meters = "meters"
    centimeters = "centimeters"
    millimeters = "millimeters"
    feet = "feet"
    inch = "inch"
    miles = "miles"


class UnitTypesArea(str, Enum):
    """Units of area.

    Attributes
    ----------
    m2 : square meters
    dm2 : square decimeters
    cm2 : square centimeters
    mm2 : square millimeters
    sf : square feet
    """

    m2 = "m2"
    dm2 = "dm2"
    cm2 = "cm2"
    mm2 = "mm2"
    sf = "sf"


class UnitTypesVolume(str, Enum):
    """Units of volume.

    Attributes
    ----------
    m3 : cubic meters
    cf : cubic feet
    """

    m3 = "m3"
    cf = "cf"


class UnitTypesVelocity(str, Enum):
    """Units of velocity.

    Attributes
    ----------
    mps : meters per second
    knots : nautical miles per hour
    mph : miles per hour
    """

    mps = "m/s"
    knots = "knots"
    mph = "mph"


class UnitTypesDirection(str, Enum):
    """Units of direction.

    Attributes
    ----------
    degrees : degrees
    """

    degrees = "deg N"


class UnitTypesTime(str, Enum):
    """Units of time.

    Attributes
    ----------
    seconds : seconds
    minutes : minutes
    hours : hours
    days : days
    """

    seconds = "seconds"
    minutes = "minutes"
    hours = "hours"
    days = "days"


class UnitTypesDischarge(str, Enum):
    """Units of discharge.

    Attributes
    ----------
    cfs : cubic feet per second
    cms : cubic meters per second
    """

    cfs = "cfs"
    cms = "m3/s"


class UnitTypesIntensity(str, Enum):
    """Units of intensity.

    Attributes
    ----------
    inch_hr : inch per hour
    mm_hr : millimeter per hour
    """

    inch_hr = "inch/hr"
    mm_hr = "mm/hr"


class VerticalReference(str, Enum):
    """Vertical reference for height.

    Attributes
    ----------
    floodmap : Use the floodmap as reference.
    datum : Use the datum as reference.
    """

    floodmap = "floodmap"
    datum = "datum"


class UnitfulLength(ValueUnitPair[UnitTypesLength]):
    """Combination of length and unit.

    Attributes
    ----------
    value : float
        The length value.
    units : UnitTypesLength
        The unit of length.
    """

    _CONVERSION_FACTORS: dict[UnitTypesLength, float] = {
        UnitTypesLength.meters: 1.0,
        UnitTypesLength.centimeters: 100.0,
        UnitTypesLength.millimeters: 1000.0,
        UnitTypesLength.feet: 3.28084,
        UnitTypesLength.inch: 1.0 / 0.0254,
        UnitTypesLength.miles: 1 / 1609.344,
    }
    _DEFAULT_UNIT: UnitTypesLength = UnitTypesLength.meters


class UnitfulHeight(UnitfulLength):
    """Combination of height and unit.

    Attributes
    ----------
    value : float
        The height value, must be greater than or equal to 0.
    units : UnitTypesLength
        The unit of height.
    """

    value: float = Field(ge=0.0)


class UnitfulLengthRefValue(UnitfulLength):
    """Combination of length and unit with a reference value.

    Attributes
    ----------
    value : float
        The length value, must be greater than or equal to 0.
    units : UnitTypesLength
        The unit of length.
    type : VerticalReference
        The vertical reference for the length.
    """

    type: VerticalReference


class UnitfulArea(ValueUnitPair[UnitTypesArea]):
    """Combination of area and unit.

    Attributes
    ----------
    value : float
        The area value, must be greater than or equal to 0.
    units : UnitTypesArea
            The unit of area.

    """

    _CONVERSION_FACTORS: dict[UnitTypesArea, float] = {
        UnitTypesArea.m2: 1,
        UnitTypesArea.dm2: 100,
        UnitTypesArea.cm2: 10_000,
        UnitTypesArea.mm2: 1_000_000,
        UnitTypesArea.sf: 10.764,
    }
    _DEFAULT_UNIT: UnitTypesArea = UnitTypesArea.m2
    value: float = Field(ge=0.0)


class UnitfulVelocity(ValueUnitPair[UnitTypesVelocity]):
    """Combination of velocity and unit.

    Attributes
    ----------
    value : float
        The velocity value, must be greater than or equal to 0.
    units : UnitTypesVelocity
        The unit of velocity.
    """

    _CONVERSION_FACTORS: dict[UnitTypesVelocity, float] = {
        UnitTypesVelocity.mph: 2.236936,
        UnitTypesVelocity.mps: 1,
        UnitTypesVelocity.knots: 1.943844,
    }
    _DEFAULT_UNIT: UnitTypesVelocity = UnitTypesVelocity.mps
    value: float = Field(ge=0.0)


class UnitfulDirection(ValueUnitPair[UnitTypesDirection]):
    """Combination of direction and unit.

    Attributes
    ----------
    value : float
        The direction value, must be greater than or equal to 0 and less than or equal to 360.
    units : UnitTypesDirection
        The unit of direction.
    """

    _CONVERSION_FACTORS: dict[UnitTypesDirection, float] = {
        UnitTypesDirection.degrees: 1.0,
    }
    _DEFAULT_UNIT: UnitTypesDirection = UnitTypesDirection.degrees

    value: float = Field(ge=0.0, le=360.0)


class UnitfulDischarge(ValueUnitPair[UnitTypesDischarge]):
    """Combination of discharge and unit.

    Attributes
    ----------
    value : float
        The discharge value, must be greater than or equal to 0.
    units : UnitTypesDischarge
            The unit of discharge.
    """

    _CONVERSION_FACTORS: dict[UnitTypesDischarge, float] = {
        UnitTypesDischarge.cfs: 35.314684921034,
        UnitTypesDischarge.cms: 1,
    }
    _DEFAULT_UNIT: UnitTypesDischarge = UnitTypesDischarge.cms

    value: float = Field(ge=0.0)


class UnitfulIntensity(ValueUnitPair[UnitTypesIntensity]):
    """Combination of intensity and unit.

    Attributes
    ----------
    value : float
        The intensity value, must be greater than or equal to 0.
    units : UnitTypesIntensity
        The unit of intensity.
    """

    _CONVERSION_FACTORS: dict[UnitTypesIntensity, float] = {
        UnitTypesIntensity.inch_hr: 1 / 25.39544832,
        UnitTypesIntensity.mm_hr: 1,
    }
    _DEFAULT_UNIT: UnitTypesIntensity = UnitTypesIntensity.mm_hr
    value: float = Field(ge=0.0)


class UnitfulVolume(ValueUnitPair[UnitTypesVolume]):
    """Combination of volume and unit.

    Attributes
    ----------
    value : float
        The volume value, must be greater than or equal to 0.
    units : UnitTypesVolume
            The unit of volume.
    """

    _CONVERSION_FACTORS: dict[UnitTypesVolume, float] = {
        UnitTypesVolume.m3: 1.0,
        UnitTypesVolume.cf: 35.3146667,
    }
    _DEFAULT_UNIT: UnitTypesVolume = UnitTypesVolume.m3

    value: float = Field(ge=0.0)


class UnitfulTime(ValueUnitPair[UnitTypesTime]):
    """Combination of time and unit.

    Attributes
    ----------
    value : float
        The time value.
    units : UnitTypesTime
        The unit of time.
    """

    _CONVERSION_FACTORS: dict[UnitTypesTime, float] = {
        UnitTypesTime.days: 1.0 / 24.0,
        UnitTypesTime.hours: 1.0,
        UnitTypesTime.minutes: 60.0,
        UnitTypesTime.seconds: 60.0 * 60.0,
    }
    _DEFAULT_UNIT: UnitTypesTime = UnitTypesTime.hours

    @staticmethod
    def from_timedelta(td: timedelta) -> "UnitfulTime":
        """Convert given timedelta to UnitfulTime object."""
        return UnitfulTime(value=td.total_seconds(), units=UnitTypesTime.seconds)

    def to_timedelta(self) -> timedelta:
        """Convert given time to datetime.deltatime object, relative to UnitfulTime(0, Any).

        Returns
        -------
        datetime.timedelta
            datetime.timedelta object with representation: (days, seconds, microseconds)
        """
        return timedelta(seconds=self.convert(UnitTypesTime.seconds))


ValueUnitPairs = Union[
    UnitfulLength,
    UnitfulTime,
    UnitfulDischarge,
    UnitfulDirection,
    UnitfulVelocity,
    UnitfulIntensity,
    UnitfulArea,
    UnitfulVolume,
]

UNIT_TO_CLASS: dict[enum.EnumMeta, Type[ValueUnitPairs]] = {
    UnitTypesLength: UnitfulLength,
    UnitTypesTime: UnitfulTime,
    UnitTypesDischarge: UnitfulDischarge,
    UnitTypesDirection: UnitfulDirection,
    UnitTypesVelocity: UnitfulVelocity,
    UnitTypesIntensity: UnitfulIntensity,
    UnitTypesArea: UnitfulArea,
    UnitTypesVolume: UnitfulVolume,
}
